/*
 *	swDraw.c
 *
 *	Last modified: <2017-05-11 11:49:31 hideki>
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBaseLib.h>
#include <sp/spAudioLib.h>
#include <sp/spLib.h>
#include <sp/spComponentLib.h>

#include "swWaveAudio.h"

#include "swWindow.h"
#include "swDraw.h"
#include "swCursor.h"
#include "swEdit.h"
#include "swDialog.h"
#include "swLabel.h"
#include "swLabelDialog.h"
#include "swLabelList.h"
#include "swAnalysis.h"
#if defined(SW_SUPPORT_MORPHING)
#include "swMorphingDraw.h"
#endif
#if defined(SW_AH_CUSTOM)
#include "swWindowAH.h"
#include "swLabelAH.h"
#include "swDrawAH.h"
#endif

#define SW_DRAW_LABEL_CANVAS	/* cursor will be sometimes slow */
/*#undef SW_DRAW_LABEL_CANVAS*/ /* dragging a label will be slow */

#define SW_METER_FONT "-*-*-medium-r-normal--8-*-*-*-*-*-*-*"

swGraphics sw_graphics = NULL;

void swLockGraphicsMutex(void)
{
    if (sw_graphics == NULL) return;
    
    spLockMutex(sw_graphics->mutex);
    
    return;
}

void swUnlockGraphicsMutex(void)
{
    if (sw_graphics == NULL) return;
    
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}

void swSetColor(swConfig config)
{
    if (sw_graphics == NULL) {
	sw_graphics = xalloc(1, struct _swGraphics);
	sw_graphics->gx_fg = NULL;
	sw_graphics->gx_bg = NULL;
	sw_graphics->gx_xor = NULL;
	sw_graphics->gx_pointer = NULL;
	sw_graphics->gx_string = NULL;
	sw_graphics->gx_label = NULL;
	sw_graphics->gx_region = NULL;
	sw_graphics->gx_region_bg = NULL;
	sw_graphics->gx_scale = NULL;
	sw_graphics->gx_meter = NULL;
	sw_graphics->gx_shading = NULL;
	sw_graphics->gx_overview_xor = NULL;
	
	sw_graphics->region_pixel = 0L;
	sw_graphics->current_region_pixel = 0L;
	sw_graphics->region_bg_pixel = 0L;
	sw_graphics->region_pixel_incr = 0;
	sw_graphics->mutex = spCreateMutex(NULL);
    }
    
    spLockMutex(sw_graphics->mutex);
    
    spDebug(40, "swCreateGraphics", "%s %s %s %s\n",
	    config->wave_fg, config->wave_bg, config->pointer_color, config->label_color);
    
    if (sw_graphics->gx_fg == NULL) {
	sw_graphics->gx_fg = spCreateGraphics("Foreground",
					      SppForeground, config->wave_fg,
					      SppFontName, config->canvas_font,
					      NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_fg,
			    SppForeground, config->wave_fg,
			    SppFontName, config->canvas_font,
			    NULL);
    }
    
    if (sw_graphics->gx_bg == NULL) {
	sw_graphics->gx_bg = spCreateGraphics("Background",
					      SppForeground, config->wave_bg,
					      NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_bg,
			    SppForeground, config->wave_bg,
			    NULL);
    }
    
    if (sw_graphics->gx_pointer == NULL) {
	sw_graphics->gx_pointer = spCreateGraphics("Pointer",
						   SppForeground, config->pointer_color,
						   NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_pointer,
			    SppForeground, config->pointer_color,
			    NULL);
    }
    
    if (sw_graphics->gx_string == NULL) {
	sw_graphics->gx_string = spCreateGraphics("String",
						  SppForeground, config->string_color,
						  SppFontName, config->canvas_font,
						  NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_string,
			    SppForeground, config->string_color,
			    SppFontName, config->canvas_font,
			    NULL);
    }
    
    if (sw_graphics->gx_label == NULL) {
	sw_graphics->gx_label = spCreateGraphics("Label",
						 SppForeground, config->label_color,
						 SppFontName, config->canvas_font,
						 NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_label,
			    SppForeground, config->label_color,
			    SppFontName, config->canvas_font,
			    NULL);
    }
    {
	int i;
	spPixel pixel;
	int r, g, b;
	spGraphicsMode region_bg_mode;
	spGraphicsMode xor_mode;
	spGraphicsMode overview_xor_mode;
	char *xor_color_name;
	
	pixel = spGetColorPixel(config->wave_bg);
	r = spGetRValue(pixel);
	g = spGetGValue(pixel);
	b = spGetBValue(pixel);
	
	if (r + g + b >= /*386*/192) {
	    if (config->toplevel->graphics_mode_caps & SP_GRAPHICS_MODE_CAPS_PLUS_DARKER) {
		xor_mode = SP_GM_PLUS_DARKER;
	    } else {
		xor_mode = SP_GM_MULTIPLY;
	    }
	    xor_color_name = /*"#808080"*//*"#a8cdf1"*/"#a0c6ff"/*"#b5d5ff"*/;
	} else {
	    xor_mode = SP_GM_PLUS_LIGHTER/*SP_GM_SCREEN*/;
	    xor_color_name = "#304080"/*"#a8cdf1"*//*"#b5d5ff"*/;
	}
	
	if (r + g + b >= /*386*/192) {
	    if (config->toplevel->graphics_mode_caps & SP_GRAPHICS_MODE_CAPS_AND) {
		region_bg_mode = SP_GM_AND;
		r = MAX(r, 128);
		g = MAX(g, 128);
		b = MAX(b, 128);
	    } else if (config->toplevel->graphics_mode_caps & SP_GRAPHICS_MODE_CAPS_PLUS_DARKER) {
		region_bg_mode = SP_GM_PLUS_DARKER;
		r = g = b = 255;
	    } else {
		region_bg_mode = SP_GM_MULTIPLY;
		r = g = b = 255;
	    }
	    sw_graphics->region_pixel_incr = SW_REGION_PIXEL_NEGA_INCR;
	} else {
	    if (config->toplevel->graphics_mode_caps & SP_GRAPHICS_MODE_CAPS_AND) {
		region_bg_mode = SP_GM_OR;
		r = MIN(r, 128);
		g = MIN(g, 128);
		b = MIN(b, 128);
	    } else {
		region_bg_mode = SP_GM_PLUS_LIGHTER;
		r = b = 8;
		g = 0;
	    }
	    sw_graphics->region_pixel_incr = SW_REGION_PIXEL_POSI_INCR;
	}
	
	overview_xor_mode = SP_GM_COPY;
	
	if (config->toplevel->graphics_mode_caps & SP_GRAPHICS_MODE_CAPS_XOR) {
	    xor_mode = SP_GM_XOR;
	    xor_color_name = "max";
	    overview_xor_mode = SP_GM_XOR;
	}
	spDebug(80, "swSetColor", "xor_mode = %d, overview_xor_mode = %d\n",
		xor_mode, overview_xor_mode);

	for (i = 0;; i++) {
	    r = MAX(r + 2 * (i + 1) * sw_graphics->region_pixel_incr, 0); r = MIN(r, 255);
	    g = MAX(g + 2 * (i + 1) * sw_graphics->region_pixel_incr, 0); g = MIN(g, 255);
	    b = MAX(b + 2 * (i + 1) * sw_graphics->region_pixel_incr, 0); b = MIN(b, 255);
	    pixel = spRGB(r, g, b);

	    if (i == 0) {
		sw_graphics->region_pixel = pixel;
		spGetColorName(pixel, sw_graphics->region_color);

		sw_graphics->region_bg_pixel = pixel;
		spGetColorName(sw_graphics->region_bg_pixel, sw_graphics->region_bg_color);
	    } else if (i == 1) {
		spGetColorName(pixel, sw_graphics->region_line_color);
	    } else {
		spGetColorName(pixel, sw_graphics->region_label_color);
		break;
	    }
	}
	
	if (sw_graphics->gx_region == NULL) {
	    sw_graphics->gx_region = spCreateGraphics("Region",
						      SppForeground, sw_graphics->region_color,
						      NULL);
	} else {
	    spSetGraphicsParams(sw_graphics->gx_region,
				SppForeground, sw_graphics->region_color,
				NULL);
	}
	if (sw_graphics->gx_region_bg == NULL) {
	    sw_graphics->gx_region_bg = spCreateGraphics("RegionBG",
							 SppGraphicsMode, region_bg_mode,
							 SppForeground, sw_graphics->region_bg_color,
							 NULL);
	} else {
	    spSetGraphicsParams(sw_graphics->gx_region_bg,
				SppGraphicsMode, region_bg_mode,
				SppForeground, sw_graphics->region_bg_color,
				NULL);
	}

	if (sw_graphics->gx_xor == NULL) {
	    sw_graphics->gx_xor = spCreateGraphics("XOR",
						   SppForeground, xor_color_name,
						   SppGraphicsMode, xor_mode,
						   NULL);
	} else {
	    spSetGraphicsParams(sw_graphics->gx_xor,
				SppForeground, xor_color_name,
				SppGraphicsMode, xor_mode,
				NULL);
	}
	if (sw_graphics->gx_overview_xor == NULL) {
	    sw_graphics->gx_overview_xor = spCreateGraphics("OverviewXOR",
							    SppForeground, xor_color_name,
							    SppGraphicsMode, overview_xor_mode,
							    NULL);
	} else {
	    spSetGraphicsParams(sw_graphics->gx_overview_xor,
				SppForeground, xor_color_name,
				SppGraphicsMode, overview_xor_mode,
				NULL);
	}
    }
    
    if (sw_graphics->gx_scale == NULL) {
	sw_graphics->gx_scale = spCreateGraphics("Scale",
						 SppForeground, config->scale_color,
						 SppFontName, config->canvas_font,
						 NULL);
    } else {
	spSetGraphicsParams(sw_graphics->gx_scale,
			    SppForeground, config->scale_color,
			    SppFontName, config->canvas_font,
			    NULL);
    }
    
    if (sw_graphics->gx_meter == NULL) {
	sw_graphics->gx_meter = spCreateGraphics("Meter",
						 SppFontName, SW_METER_FONT,
						 NULL);
    }
    if (sw_graphics->gx_shading == NULL) {
	sw_graphics->gx_shading = spCreateGraphics("Shading",
						   NULL);
    }
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}

void swGetDrawRange(double y_draw_offset, double draw_height, int channel, int *draw_min, int *draw_max)
{
    double dmin, dmax;

    dmin = (double)channel * draw_height + y_draw_offset;
    dmax = dmin + draw_height;

    if (draw_min != NULL) *draw_min = (int)spRound(dmin);
    if (draw_max != NULL) *draw_max = (int)spRound(dmax);

    return;
}

int swGetDrawWidth(swWindow window)
{
#ifdef SW_SUPPORT_METER
    return MAX(window->width - window->meter_width, SW_MIN_DRAW_WIDTH);
#else
    return MAX(window->width, SW_MIN_DRAW_WIDTH);
#endif
}

double swGetWindowLimitValue(swWindow window, swWave wave)
{
#if 0
    if (swIsWaveFloat(wave) == SP_TRUE) {
	double min, max;
	min = swGetWaveMin(wave);
	max = swGetWaveMax(wave);
	max = MAX(FABS(min), FABS(max));
	if (max < 1.0) {
	    max = 1.0;
	}
	return max;
    }
#endif
    return swGetLimitValue(wave->samp_bit);
}

spBool swGetMainStringExtent(spComponent canvas, char *string, int *x, int *y, int *width, int *height)
{
    return spGetStringExtent(canvas, sw_graphics->gx_string, string, x, y, width, height, NULL);
}

void swDrawMainString(spComponent canvas, int left_offset, int top_offset, char *string)
{
    spDrawString(canvas, sw_graphics->gx_string, left_offset, top_offset, string);
    return;
}

#ifdef SW_SUPPORT_METER
spBool swIsMeterVisible(swWindow window)
{
    if (window == NULL /*|| window->draw_height <= 0.0*/
	|| window->meter_width <= 0) {
	return SP_FALSE;
    }

    return SP_TRUE;
}

#define SW_MIN_METER_VALUE -300.0
#define SW_METER_DRAW_HEIGHT_THRESHOLD 150
#define SW_METER_TICK_STRING_LEFT_OFFSET 5
#define SW_METER_TICK_STRING_BOTTOM_OFFSET 5

void swDrawMeterBar(spComponent component, swWindow window, 
		    double draw_height, int left_offset, int top_offset, double value)
{
    int i;
    int y, py;
    int height;
    int meter_range;
    int meter_min, meter_max;
    double y_factor;
    unsigned char green, red;
    char buf[SP_MAX_LINE];

    if (swIsMeterVisible(window) == SP_FALSE) return;

    meter_min = -window->config->meter_range;
    meter_max = 6;
    meter_range = meter_max - meter_min;

    if (meter_range <= 0) return;
    
    spLockMutex(sw_graphics->mutex);
    
    y_factor = draw_height / (double)meter_range;
    spDebug(60, "swDrawMeterBar", "y_factor = %f, draw_height = %f, left_offset = %d, top_offset = %d\n",
	    y_factor, draw_height, left_offset, top_offset);

    py = top_offset + (int)draw_height;
    for (i = meter_min; i <= meter_max; i++) {
	y = top_offset + (int)draw_height - (int)spRound(y_factor * (double)(i - meter_min));
	height = py - y;

	if (height >= 1) {
	    if (i < -24) {
		green = (unsigned char)MAX(256 + 3 * (24 + i), 0);
		red = 0;
	    } else if (i > -6) {
		green = MAX(-128 * i / 3, 0);
		red = 255;
	    } else if (i > -18) {
		green = 255;
		red = MIN(64 * i / 3 + 384, 255);
	    } else {
		green = 255;
		red = 0;
	    }
	    
	    if ((double)i > value) {
		green /= 2;
		red /= 2;
	    }
	    
	    spDebug(120, "swDrawMeterBar", "i = %d, green = %d, red = %d, y = %d, height = %d\n",
		    i, green, red, y, height);
	    spSetForegroundPixel(sw_graphics->gx_meter, spRGB(red, green, 0));
	    
	    spFillRectangle(component, sw_graphics->gx_meter,
			    left_offset, y,
			    window->meter_width, height);
	    
	    py = y;
	}
    }
    
    py = top_offset + (int)draw_height;
    for (i = meter_min; i < meter_max; i++) {
	y = top_offset + (int)draw_height - (int)spRound(y_factor * (double)(i - meter_min));
	if ((draw_height > SW_METER_DRAW_HEIGHT_THRESHOLD && i >= -6 && i % 3 == 0)
	    || (i >= -12 && i % 6 == 0)
	    || (i % 24 == 0)) {
	    /* draw tick */
	    spSetForegroundPixel(sw_graphics->gx_meter, spRGB(255, 255, 255));
	    
	    spDrawLine(component, sw_graphics->gx_meter,
		       left_offset, y, left_offset + 2, y);
	    spDrawLine(component, sw_graphics->gx_meter,
		       left_offset + window->meter_width,
		       y, left_offset + window->meter_width - 3, y);

	    sprintf(buf, "%3d", i);
	    spDrawString(component, sw_graphics->gx_meter,
			 left_offset + SW_METER_TICK_STRING_LEFT_OFFSET,
			 y + SW_METER_TICK_STRING_BOTTOM_OFFSET, buf);
	}
    }
    
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}

void swDrawMeter(spComponent component, swWindow window, swWave wave,
		 double y_draw_offset, double draw_height, int left_offset, int channel, double value)
{
    double max;
    double dbvalue;
    
    if (swIsMeterVisible(window) == SP_FALSE) return;

    max = swGetWindowLimitValue(window, wave);

    dbvalue = FABS(value) / max;
    dbvalue = (dbvalue < 1e-30 ? -300.0 : dB(dbvalue));
    
    spDebug(80, "swDrawMeter", "channel = %d, value = %f, dbvalue = %f, max = %f\n",
	    channel, value, dbvalue, max);
    swDrawMeterBar(component, window, draw_height, left_offset,
		   (int)spRound((double)channel * draw_height + y_draw_offset),
		   dbvalue);
    
    return;
}
#endif

double swGetDataRange(swWindow window, swWave wave, double *data_min, double *data_max)
{
    double dmin, dmax;

    if (wave->num_order > 1) {
	dmin = 0.0;
	dmax = (double)wave->num_order;
    } else {
	dmin = swGetWaveMin(wave);
	dmax = swGetWaveMax(wave);
	spDebug(100, "swGetDataRange", "dmin = %f, dmax = %f\n", dmin, dmax);
    }
    
    if (data_min != NULL) *data_min = dmin;
    if (data_max != NULL) *data_max = dmax;

    return dmax - dmin;
}

double swGetAmplitudeRange(swWindow window, swWave wave, double *data_min, double *data_max)
{
    double dmin, dmax;

    if (window->amp_max <= window->amp_min) {
	return swGetDataRange(window, wave, data_min, data_max);
    }

    dmin = window->amp_min;
    dmax = window->amp_max;
    
    if (data_min != NULL) *data_min = dmin;
    if (data_max != NULL) *data_max = dmax;

    return dmax - dmin;
}

double swGetOrderRange(swWindow window, swWave wave, double *p_order_min, double *p_order_max)
{
    double amp_min, amp_max;
    double order_min;
    double order_max;
    double order_range;
    
    swGetAmplitudeRange(window, wave, &amp_min, &amp_max);
    order_min = MAX(amp_min, 0.0);
    order_max = MIN(amp_max, (double)wave->num_order);
    order_max = MAX(order_max, 0.0);
    order_range = MIN((double)wave->num_order, order_max - order_min);
    order_range = MAX(order_range, 1.0);
    
    if (p_order_min != NULL) *p_order_min = order_min;
    if (p_order_max != NULL) *p_order_max = order_max;

    return order_range;
}

double swGetDataHeight(swWindow window, swWave wave, double *data_min, double *data_max)
{
    double dmin, dmax;
    
    swGetDataRange(window, wave, &dmin, &dmax);
    
    if (dmin >= dmax) {
	dmin = 0.0; dmax = 1.0;
    }
    
    if (data_min != NULL) *data_min = dmin;
    if (data_max != NULL) *data_max = dmax;

    return dmax - dmin;
}

static double swGetScaleTick(int width, double data_width, double *pw, spBool v_flag)
{
    int idx;
    double ndiv;
    double pn, fract;
    double tick, ntick;
    /* table for drawing scale */
    static double scale_table[] = {1.0, 2.0, 5.0, 10.0};
    static double scale_log_table[] = {0.0, 0.301, 0.699, 1.0};

    if (data_width <= 0.0) {
	fract = 0.0;
	pn = 0.0;
    } else {
	if (v_flag == SP_TRUE) {
	    ndiv = MAX(floor((double)width / SW_MIN_VSCALE_DIV), 2.0);
	} else {
	    ndiv = MAX(floor((double)width / SW_MIN_HSCALE_DIV), 2.0);
	}
	fract = modf(log10(data_width / ndiv), &pn);
    }

    if (pn < 0.0 || fract < 0.0) {
	pn = pn - 1.0;
	fract = 1.0 + fract;
    }

    spDebug(40, "swGetScaleTick", "pn = %f, fract = %f\n", pn, fract);

    if (fract > scale_log_table[0] && fract <= scale_log_table[1]) {
	idx = 1;
    } else if (fract > scale_log_table[1] && fract <= scale_log_table[2]) {
	idx = 2;
    } else if (fract > scale_log_table[2] && fract <= scale_log_table[3]) {
	idx = 0;
	pn++;
    } else {
	idx = 0;
    }
    tick = scale_table[idx];

    ntick =  tick * pow(10.0, pn);
    ndiv = floor(data_width / ntick);
    if (ndiv * tick >= 10.0) {
	tick /= 10.0;
	pn++;
    }

    if (pw != NULL) {
	*pw = pn;
    }

    return tick;
}

int swGetScaleString(char *string, double tick, double pn, double value)
{
    int flag = 0;
    int pni;
    double n, ntick;
    char exp_string[SP_MAX_MESSAGE];

    n = pow(10.0, pn);
    ntick = n * tick;
    
    if (value == 0.0) {
	sprintf(string, "0");
    } else if (pn < 0) {
	pni = (int)spRound(pn);
	if (tick < 1.0) {
	    pni--;
	}
	switch (pni) {
	  case -1:
	    sprintf(string, "%.1f", value);
	    break;
	  case -2:
	    sprintf(string, "%.2f", value);
	    break;
	  case -3:
	    sprintf(string, "%.3f", value);
	    break;
	  case -4:
	    sprintf(string, "%.4f", value);
	    break;
	  case -5:
	    sprintf(string, "%.5f", value);
	    break;
	  default:
	    if (tick < 1.0) {
		sprintf(string, "%.1f", value / n);
	    } else {
		sprintf(string, "%.0f", value / n);
	    }
	    flag = 1;
	}
    } else {
	if (pn >= 6.0) {
	    if (tick < 1.0) {
		sprintf(string, "%.1f", value / n);
	    } else {
		sprintf(string, "%.0f", value / n);
	    }
	    flag = 1;
	} else {
	    if (ntick < 1.0) {
		sprintf(string, "%.1f", value);
	    } else {
		sprintf(string, "%.0f", value);
	    }
	}
    }
		
    if (flag) {
	sprintf(exp_string, "x10e%.0f", pn);
	strcat(string, exp_string);
    }

    return flag;
}

static void swClipZeroChars(char *string)
{
    int i;
    
    for (i = (int)strlen(string) - 1; i >= 0; i--) {
	if (string[i] == '0') {
	    string[i] = NUL;
	} else {
	    if (i >= 1 && string[i] == '.') {
		string[i] = NUL;
	    }
	    break;
	}
    }

    return;
}

#if 1
void swDrawHScaleMain(swConfig config, spComponent component, swDataType data_type,
		      double y_draw_offset, double draw_width, double draw_height, 
		      int draw_min, int draw_max, double x_min, double x_max, int x_width,
		      int num_channel, int channel, spBool for_print)
{
    int w, h;
    int x, y;
    int strx, stry;
    int max_str_height;
    int x_offset;
    int direct;
    long k;
    long ndmax, ndmin;
    double tick, ntick;
    double n, pn;
    double x_data_width;
    double x_zero, x_fact;
    double sx;
    char string[SP_MAX_MESSAGE];

    direct = 1;
    x_data_width = x_max - x_min;

    if (for_print == SP_TRUE) {
	x_offset = SW_PRINT_LEFT_MARGIN;
    } else {
	x_offset = 0;
    }

    tick = swGetScaleTick(x_width, x_data_width, &pn, SP_FALSE);
    n = pow(10.0, pn);
    ntick = n * tick;

    ndmin = (long)floor(x_min / ntick);
    ndmax = (long)ceil(x_max / ntick);

    x_fact = (double)(direct * x_width) / x_data_width;
    x_zero = x_fact * (direct > 0 ? -x_min : x_max);

#if 1
    if (channel <= 0 && draw_min > 0) {
	spDrawLine(component, sw_graphics->gx_scale, x_offset, draw_min,
		   x_offset + (int)spRound(draw_width), draw_min);
    }
#endif
    
    if (config->grid_flag == SP_TRUE) {
	spLockMutex(sw_graphics->mutex);
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_DOT, NULL);
	for (k = ndmin; k <= ndmax; k++) {
	    sx = (double)k * ntick;
	    x = (int)spRound(x_zero + sx * x_fact) + x_offset;

	    /*if (x >= x_offset && x <= x_offset + x_width) {*/
	    if (x > x_offset && x < x_offset + x_width) {
		/*if (x == x_width) x -= 1;*/
		
		spDrawLine(component, sw_graphics->gx_scale, x, draw_min, x, draw_max);
	    }
	}
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_SOLID, NULL);
	spUnlockMutex(sw_graphics->mutex);
    }

    if (draw_max <= draw_min + (int)spRound(draw_height)) {
	spDrawLine(component, sw_graphics->gx_scale, x_offset, draw_max,
		   x_offset + (int)spRound(draw_width), draw_max);
    } else {
	spDebug(40, "swDrawHScaleMain", "**** draw_min = %d, draw_max = %d, draw_height = %f\n",
		draw_min, draw_max, draw_height);
    }
    
    if (config->scale_flag == SP_FALSE) {
	return;
    }

    spLockMutex(sw_graphics->mutex);
    
    if (for_print == SP_TRUE) {
	spSetGraphicsParams(sw_graphics->gx_scale,
			    SppForeground, "black",
			    NULL);
    }

    max_str_height = 0;
    
    for (k = ndmin; k <= ndmax; k++) {
	sx = (double)k * ntick;
	x = (int)spRound(x_zero + sx * x_fact) + x_offset;

	if (x >= x_offset && x <= x_offset + x_width) {
	    /*if (x == x_width) x -= 1;*/
	    
	    if (x > x_offset && x < x_offset + x_width) {
		spDrawLine(component, sw_graphics->gx_scale, x, draw_min, x, draw_min + SW_TICK_LENGTH);
		spDrawLine(component, sw_graphics->gx_scale, x, draw_max, x, draw_max - SW_TICK_LENGTH);
	    }

	    if ((x >= SW_HSCALE_LEFT_SPACING || for_print == SP_TRUE)
		&& channel == num_channel - 1) {
		if (data_type != SW_FREQ_DATA
		    && config->time_format == SW_TIME_FORMAT_SEPARATED_SEC) {
		    spGetTimeString(sx, SP_TIME_FORMAT_SEPARATED_SEC, string);
		    swClipZeroChars(string);
		} else {
		    swGetScaleString(string, tick, pn, sx);
		}
		
		if (spGetStringExtent(component, sw_graphics->gx_scale, string, &strx, &stry, &w, &h, NULL) == SP_TRUE) {
		    if (max_str_height < h) max_str_height = h;
		    
		    /*x = x - w / 2;*/
		    if (for_print == SP_TRUE) {
			y = draw_max - stry + SW_PRINT_HSCALE_STRING_SPACING;
		    } else {
			x = MAX(x, SW_HSCALE_STRING_LEFT_OFFSET);
			y = draw_max - SW_TICK_LENGTH - SW_HSCALE_STRING_BOTTOM_OFFSET;
		    }
		    spSetGraphicsParams(sw_graphics->gx_scale,
					SppStringAlignment, SP_ALIGNMENT_CENTER,
					NULL);
		    spDrawString(component, sw_graphics->gx_scale, x, y, string);
		    spSetGraphicsParams(sw_graphics->gx_scale,
					SppStringAlignment, SP_ALIGNMENT_BEGINNING,
					NULL);
		}
	    }
	}
    }

    if (for_print == SP_TRUE && channel == num_channel - 1) {
        int y_offset;
	char title_string[SP_MAX_LINE];
	char dim_string[SP_MAX_LINE];
	
	spSetGraphicsParams(sw_graphics->gx_scale,
			    SppLineWidth, 2,
			    SppStringAlignment, SP_ALIGNMENT_CENTER,
			    NULL);
        y_offset = (int)spRound(y_draw_offset);
	spDrawRectangle(component, sw_graphics->gx_scale, x_offset, y_offset, x_width, draw_max - y_offset);

	swGetTimeStringTitle(config, data_type, title_string, dim_string);
	sprintf(string, "%s %s", title_string, dim_string);
	
	if (spGetStringExtent(component, sw_graphics->gx_scale, string, &strx, &stry, &w, &h, NULL) == SP_TRUE) {
	    spDrawString(component, sw_graphics->gx_scale, x_offset + x_width / 2,
			 draw_max + max_str_height + SW_PRINT_XLABEL_SPACING - stry, string);
	}
	    
	spSetGraphicsParams(sw_graphics->gx_scale,
			    SppForeground, config->scale_color,
			    SppLineWidth, 1, 
			    SppStringAlignment, SP_ALIGNMENT_BEGINNING,
			    NULL);
    }
    
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}
#endif

void swDrawHScale(spComponent component, swWindow window, swWave wave,
		  double y_draw_offset, double draw_width, double draw_height,
		  int channel, spLong offset, spLong length, spBool for_print)
{
    int x_width;
    int draw_min, draw_max;
    double x_max, x_min;

    swGetDrawRange(y_draw_offset, draw_height, channel, &draw_min, &draw_max);
    x_width = (int)spRound(draw_width);

    x_min = swSampToCurrentDim(window, offset);
    x_max = swSampToCurrentDim(window, offset + length - 1);

    swDrawHScaleMain(window->config, component, window->data_type,
		     y_draw_offset, draw_width, draw_height,
		     draw_min, draw_max, x_min, x_max, x_width, wave->num_channel, channel, for_print);

    return;
}

#if 1
void swDrawVScaleMain(swConfig config, spComponent component, swDataType data_type,
		      int direction, int x_width, int y_width, double y_draw_offset, double draw_height,
		      int num_channel, int channel, double data_min, double data_height,
		      double data_fact, double *y_factor, double *y_zero_offset,
		      char *y_label, spBool draw_flag, spBool for_print)
{
    int flag;
    int x, y;
    int x_offset;
    int direct;
    int draw_min, draw_max;
    int w, h;
    int strx, stry;
    int max_str_width;
    int v_str_offset;
    long k;
    long ndmax, ndmin;
    double data_max;
    double tick, ntick;
    double n, pn;
    double y_fact, y_zero;
    double sy;
    char string[SP_MAX_MESSAGE];

    spDebug(40, "swDrawVScaleMain", "data_min = %f, data_height = %f\n",
	    data_min, data_height);
    
    swGetDrawRange(y_draw_offset, draw_height, channel, &draw_min, &draw_max);

    flag = 0;
    direct = -1;
    data_max = data_min + data_height;
    if (data_min == data_max) {
	flag = 1;
    }

    if (for_print == SP_TRUE) {
	x_offset = SW_PRINT_LEFT_MARGIN;
    } else {
	x_offset = 0;
    }

    tick = swGetScaleTick(y_width, data_height, &pn, SP_TRUE);
    n = pow(10.0, pn);
    ntick = n * tick;

    ndmax = (long)ceil(data_max / ntick);
    ndmin = (long)floor(data_min / ntick);
    if (flag) {
	data_min = ntick * (double)ndmin;
	data_max = ntick * (double)ndmax;

	if (data_max == data_min) {
	    data_min -= 1.0;
	    data_max += 1.0;
	}
	data_height = data_max - data_min;
    }

    spDebug(40, "swDrawVScaleMain", "data_min = %f, data_max = %f\n",
	    data_min, data_max);
    
    y_fact = (double)(direction * (y_width)) / data_height;
    y_zero = y_fact * ((direction > 0) ? data_max : -data_min);

    if (y_factor != NULL) *y_factor = (100.0 / data_fact) * y_fact;
    if (y_zero_offset != NULL) *y_zero_offset = y_zero;

    if (draw_flag == SP_FALSE) return;
    
    y_zero += (double)draw_min;
    
    if (config->grid_flag == SP_TRUE || config->zero_flag == SP_TRUE) {
	spLockMutex(sw_graphics->mutex);
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_DOT, NULL);
	if (config->grid_flag == SP_TRUE) {
	    for (k = ndmin; k <= ndmax; k++) {
		sy = (double)k * ntick;
		y = (int)spRound(y_zero - sy * y_fact);

		/*if (y >= draw_min && y <= draw_max) {*/
		if (y > draw_min && y < draw_max) {
		    spDrawLine(component, sw_graphics->gx_scale, x_offset, y, x_offset + x_width, y);
		}
	    }
	} else {
	    y = (int)spRound(y_zero);
	    /*if (y >= draw_min && y <= draw_max) {*/
	    if (y > draw_min && y < draw_max) {
		spDrawLine(component, sw_graphics->gx_scale, x_offset, y, x_offset + x_width, y);
	    }
	}
	spSetGraphicsParams(sw_graphics->gx_scale, SppLineType, SP_LINE_SOLID, NULL);
	spUnlockMutex(sw_graphics->mutex);
    }

    if (config->scale_flag == SP_TRUE) {
	spLockMutex(sw_graphics->mutex);
	
	if (for_print == SP_TRUE) {
	    spSetGraphicsParams(sw_graphics->gx_scale,
				SppForeground, "black",
				SppStringAlignment, SP_ALIGNMENT_END,
				NULL);
	}

	max_str_width = 0;
	
	for (k = ndmin; k <= ndmax; k++) {
	    sy = (double)k * ntick;
	    y = (int)spRound(y_zero - sy * y_fact);

	    if (y >= draw_min && y <= draw_max) {
		if (y > draw_min && y < draw_max) {
		    spDrawLine(component, sw_graphics->gx_scale, x_offset, y, x_offset + SW_TICK_LENGTH, y);
		    spDrawLine(component, sw_graphics->gx_scale,
			       x_offset + x_width - SW_TICK_LENGTH, y, x_offset + x_width, y);
		}
	    
		swGetScaleString(string, tick, pn, sy);
	    
		if (spGetStringExtent(component, sw_graphics->gx_scale, string, &strx, &stry, &w, &h, NULL) == SP_TRUE) {
		    if (max_str_width < w) max_str_width = w;
		    
		    y = y - direct * ((-stry) / 2 - SW_VSCALE_STRING_BOTTOM_OFFSET);
		    
		    if (for_print == SP_TRUE) {
			x = x_offset - SW_PRINT_VSCALE_STRING_SPACING;
		    } else {
			x = x_offset + SW_TICK_LENGTH + SW_VSCALE_STRING_LEFT_OFFSET;
		    }

		    if (for_print == SP_FALSE || channel > 0) {
			if (for_print == SP_TRUE) {
			    v_str_offset = SW_PRINT_VSCALE_STRING_TOP_OFFSET;
			} else {
			    v_str_offset = SW_VSCALE_STRING_TOP_OFFSET;
			}
			y = MAX(y, (draw_min + v_str_offset - direct * (-stry)));
		    }
		    if (for_print == SP_FALSE || channel < num_channel - 1) {
			if (for_print == SP_TRUE) {
			    v_str_offset = SW_PRINT_VSCALE_STRING_BOTTOM_OFFSET;
			} else {
			    v_str_offset = SW_VSCALE_STRING_BOTTOM_OFFSET;
			}
			y = MIN(y, draw_max + direct * v_str_offset);
		    }
			
		    spDrawString(component, sw_graphics->gx_scale, x, y, string);
		}
	    }
	}
	
	if (for_print == SP_TRUE) {
	    if (!strnone(y_label) && channel == num_channel - 1) {
		spSetGraphicsParams(sw_graphics->gx_scale,
				    SppStringRotation, 90,
				    SppStringAlignment, SP_ALIGNMENT_CENTER,
				    NULL);

		spDrawString(component, sw_graphics->gx_scale,
			     x_offset - max_str_width - SW_PRINT_YLABEL_SPACING,
			     draw_max / 2, y_label);
	    }
	    
	    spSetGraphicsParams(sw_graphics->gx_scale,
				SppForeground, config->scale_color,
				SppStringAlignment, SP_ALIGNMENT_BEGINNING,
				SppStringRotation, 0,
				NULL);
	}
	
	spUnlockMutex(sw_graphics->mutex);
    }

    return;
}
#endif

void swDrawVScale(spComponent component, swWindow window, swWave wave,
		  double y_draw_offset, double draw_width, double draw_height,
		  int channel, double data_min, double data_height,
		  double *y_factor, double *y_zero_offset, spBool draw_flag, spBool specgram_flag, spBool for_print)
{
    int x_width;
    int y_width;
    char y_label[SP_MAX_LINE];
    double data_fact = 100.0;
    char *str;

    if (wave->num_order > 1) {
	/*if (window->specgram == wave) {*/
	if (specgram_flag == SP_TRUE) {
	    double value;

	    value = (window->wave->samp_rate / 2.0) / (double)wave->num_order;
	    data_min *= value;
	    data_height *= value;
	}
    } else {
	if (wave->num_order <= 1
	    && window->config->percent_amplitude == SP_TRUE
	    && window->data_type != SW_FREQ_DATA) {
	    data_fact = swGetWindowLimitValue(window, wave);
	
	    data_min *= (100.0 / data_fact);
	    data_height *= (100.0 / data_fact);
	}
    }

    spDebug(40, "swDrawVScale", "data_min = %f, data_height = %f\n",
	    data_min, data_height);
    
    x_width = (int)draw_width;
    y_width = (int)draw_height;

    if (draw_flag == SP_TRUE) {
	if (wave->num_order > 1) {
	    strcpy(y_label, "Frequency [Hz]");
	} else {
	    str = swGetAnalysisNameString(wave);
	    if (!strnone(str)) {
		strcpy(y_label, str);
	    } else {
		strcpy(y_label, "Amplitude");
	    }
	    str = swGetAnalysisUnitString(wave);
	    if (!strnone(str)) {
		if (streq(y_label, "Amplitude") && streq(str, "[dB]")) {
		    strcpy(y_label, "Magnitude");
		}
		strcat(y_label, " ");
		strcat(y_label, str);
	    } else if (window->config->percent_amplitude == SP_TRUE) {
		strcat(y_label, " ");
		strcat(y_label, "[%]");
	    }
	}
    }
    
    swDrawVScaleMain(window->config, component, window->data_type,
		     window->direction, x_width, y_width, y_draw_offset, draw_height,
		     wave->num_channel, channel, data_min, data_height,
		     data_fact, y_factor, y_zero_offset,
		     y_label, draw_flag, for_print);
    
    return;
}

static void drawCurrentLine(spComponent component, spGraphics graphics,
			    spLong thin_length, int draw_min, int draw_max,
			    int pmin, int pmax, int min, int max, int px, int py, int x, int y)
{
    spBool rect_flag = SP_FALSE;

    if (max < draw_min || min > draw_max) {
	/* outside of canvas */
    } else {
	if (x - px > 1) {
	    rect_flag = SP_TRUE;
#if 0
	    spDebug(100, "drawCurrentLine", "thin_length = %ld, x - px = %d\n",
		    thin_length, x - px);
#endif
	}
    
	if (min > pmax || max < pmin) {
	    /* draw line from previous rectangle */
	    pmin = MIN(min, pmin);
	    pmax = MAX(max, pmax);
	    pmin = MAX(pmin, draw_min);
	    pmax = MIN(pmax, draw_max);
	    if (rect_flag == SP_TRUE) {
		spDrawLine(component, sw_graphics->gx_fg, px, pmin, px, pmax);
		pmin = min; pmax = max;
	    }
	} else {
	    pmin = min; pmax = max;
	}

	pmin = MAX(pmin, draw_min);
	pmax = MIN(pmax, draw_max);
	if (rect_flag == SP_TRUE) {
	    spFillRectangle(component, graphics, px, pmin, x - px,  MAX(pmax - pmin, 1));
	} else {
	    spDrawLine(component, graphics, px, pmin, px, pmax);
	}
    }

    return;
}

#if 1
spPixel swGetShadingPixel(swWindow window, swWave wave, double value)
{
    double min, max;
    double range;

    range = window->config->specgram_range;
    if (range <= 0.0) range = 1.0;
    max = swGetWaveMax(wave) + (double)MIN(0, window->config->specgram_limit_threshold);
    min = max - range;
    
    if (window->config->specgram_gray_scale == SP_TRUE) {
	unsigned char r, g, b;
	
	value = 255.0 * (1.0 - ((value - min) / range));

	if (value >= 255.0) {
	    r = g = b = 255;
	} else if (value <= 0.0) {
	    r = g = b = 0;
	} else {
	    r = g = b = (unsigned char)round(value);
	}

	return spRGB(r, g, b);
    } else {
	double r, g, b;
	
	value = 256.0 * ((value - min) / range);

	if (value >= 256.0) {
	    r = 128.0;
	    g = b = 0;
	} else if (value >= 224.0) {
	    r = 256.0 - (4.0 * (value - 224.0));
	    g = b = 0;
	} else if (value >= 160.0) {
	    g = 256.0 - (4.0 * (value - 160.0));
	    r = 256.0;
	    b = 0;
	} else if (value >= 96.0) {
	    r = 4.0 * (value - 96.0);
	    b = 256.0 - 4.0 * (value - 96.0);
	    g = 256.0;
	} else if (value >= 32.0) {
	    g = 4.0 * (value - 32.0);
	    b = 256.0;
	    r = 0;
	} else if (value > 0.0) {
	    b = 4.0 * (value + 32.0);
	    r = g = 0;
	} else {
	    b = 128.0;
	    r = g = 0;
	}

	r = MAX(0.0, r); r = MIN(255.0, r);
	g = MAX(0.0, g); g = MIN(255.0, g);
	b = MAX(0.0, b); b = MIN(255.0, b);
	
	return spRGB((unsigned char)round(r), (unsigned char)round(g),
		     (unsigned char)round(b));
    }
}

static void swFillShadingRectangle(spComponent component, swWindow window, swWave wave,
				   int draw_min, int draw_size,
				   int x, int px, double min, double max, spLong min_index, spLong max_index,
				   spBool ignore_min, double *last_value, spLong *last_index)
{
    spPixel first_pixel, latter_pixel;
    int first_len, latter_len;
    int x_update_min_size = 2;

    if (x <= 0 || x <= px) return;

    px = MAX(px, 0);
    
    if (ignore_min == SP_TRUE || (min == max && (*last_value == max || *last_index == -1))) {
	first_pixel = swGetShadingPixel(window, wave, max);
	spSetForegroundPixel(sw_graphics->gx_shading, first_pixel);
	spFillRectangle(component, sw_graphics->gx_shading, px, draw_min, x - px, draw_size);
	*last_index = MAX(min_index, max_index);
	*last_value = max;
    } else {
	int i;
	int num_loop;
	int pos;
	int remain_size;
	int current_draw_size;
	double weight, weight1, weight2;
	double first_value, half_value;
	double prev_value, value;
	spPixel pixel;

	num_loop = (x - px) / x_update_min_size;

	if (num_loop < 3) {
	    /* fill rectangle in first half */
	    if (min_index < max_index) {
		first_pixel = swGetShadingPixel(window, wave, min);
		latter_pixel = swGetShadingPixel(window, wave, max);
		*last_index = max_index;
		*last_value = max;
	    } else {
		first_pixel = swGetShadingPixel(window, wave, max);
		latter_pixel = swGetShadingPixel(window, wave, min);
		*last_index = min_index;
		*last_value = min;
	    }
	    first_len = (x - px) / 2;
	    if (first_len > 0) {
		spSetForegroundPixel(sw_graphics->gx_shading, first_pixel);
		spFillRectangle(component, sw_graphics->gx_shading, px, draw_min, first_len, draw_size);
	    }

	    /* fill rectangle in latter half */
	    latter_len = (x - px) - first_len;
	    if (latter_len > 0) {
		spSetForegroundPixel(sw_graphics->gx_shading, latter_pixel);
		spFillRectangle(component, sw_graphics->gx_shading, px + first_len, draw_min, latter_len, draw_size);
	    }
	} else {
#if 0
	    if (px < 30 && min_index != max_index) {
		spDebug(80, "swFillShadingRectangle", "num_loop = %d, x = %d, px = %d, last_index = %ld, last_value = %f\n",
			num_loop, x, px, *last_index, *last_value);
	    }
#endif
	    
	    if (*last_index < 0) {
		if (min_index < max_index) {
		    first_value = min;
		} else {
		    first_value = max;
		}
	    } else {
		first_value = *last_value;
	    }
	    
	    if (min_index < max_index) {
		half_value = min;
		*last_index = max_index;
		*last_value = max;
	    } else {
		if (min_index == max_index) {
		    half_value = (first_value + max) / 2.0;
		} else {
		    half_value = max;
		}
		*last_index = min_index;
		*last_value = min;
	    }
#if 0
	    if (px < 30 && min_index != max_index) {
		spDebug(80, "swFillShadingRectangle",
			"min_index = %ld, max_index = %ld, first_value = %f half_value = %f, last_value = %f\n",
			min_index, max_index, first_value, half_value, *last_value);
	    }
#endif

	    weight1 = (half_value - first_value) / (double)(num_loop / 2);
	    weight2 = (*last_value - half_value) / (double)(num_loop / 2);

	    pos = px;
	    remain_size = x - px;
	    prev_value = first_value;
	
	    for (i = 0; i < num_loop; i++) {
		if (remain_size <= 0) {
		    break;
		}

		if (i >= num_loop - 1) {
		    current_draw_size = remain_size;
		} else {
		    current_draw_size = x_update_min_size;
		}

		if (i < num_loop / 2) {
		    weight = weight1;
		} else {
		    weight = weight2;
		}
	    
		value = prev_value + weight;

#if 0
		if (px < 30 && min_index != max_index && (i == num_loop / 2 - 1 || i == num_loop - 1)) {
		    spDebug(80, "swFillShadingRectangle",
			    "i = %d, value = %f, first_value = %f, half_value = %f, last_value = %f, weight = %f\n",
			    i, value, first_value, half_value, *last_value, weight, value);
		}
#endif

		pixel = swGetShadingPixel(window, wave, value);
		spSetForegroundPixel(sw_graphics->gx_shading, pixel);
		spFillRectangle(component, sw_graphics->gx_shading, pos, draw_min, current_draw_size, draw_size);

		remain_size -= current_draw_size;
		pos += current_draw_size;

		prev_value = value;
	    }
	}
    }

    return;
}

void swDrawWaveformShading(spComponent component, swWindow window, swWave wave, double y_draw_offset,
			   double draw_width, double draw_height,
			   int x_offset, double x_factor, double y_factor,
			   int channel, spLong offset, spLong length,
			   spLong draw_offset, spLong draw_length, spBool peak_flag, spBool for_print)
{
    int x, px;
    int x_update_size;
    int draw_min, draw_max;
    int draw_width_i;
    double draw_size;
    spLong k, l;
    long m;
    spLong thin_length;
    spLong wave_offset;
    double value;
    spLong min_index, max_index;
    double min, max;
    spLong last_index;
    double last_value;
    double x_factor2;
    double simplify_thresh;
    long order, prev_order;
    double order_min, order_max, order_range;
    double length_ms;
    int order_draw_min, order_draw_max;
    int order_draw_size;
    int order_draw_size_min;
    spBool ignore_min;
    spBool flag;

    swGetDrawRange(y_draw_offset, draw_height, channel, &draw_min, &draw_max);
    
    spDebug(50, "swDrawWaveformShading", "x_factor = %f, y_factor = %f, draw_offset = %ld, draw_length = %ld\n",
	    x_factor, y_factor, draw_offset, draw_length);

    spLockMutex(sw_graphics->mutex);

    order_range = swGetOrderRange(window, wave, &order_min, &order_max);
    draw_size = (draw_max - draw_min) / order_range;
    
    draw_width_i = (int)round(draw_width);
    spDebug(50, "swDrawWaveformShading", "order_min = %f, order_max = %f, order_range = %f\n",
	    order_min, order_max, order_range);
    
    if (peak_flag == SP_TRUE) {
	thin_length = wave->peak_buf_thin_length;
    } else {
	thin_length = wave->thin_length;
    }
    x_factor2 = x_factor / (double)thin_length;
    
    wave_offset = swGetWaveOffset(wave);
    spDebug(50, "swDrawWaveformShading", "draw_size = %f, thin_length = %ld, play_flag = %d, process_flag = %d\n",
	    draw_size, thin_length, window->wave->core->play_flag, window->wave->core->process_flag);

    if (window->draw_detail == SP_TRUE || for_print == SP_TRUE) {
	simplify_thresh = 2000.0;
    } else {
	simplify_thresh = 1000.0;
    }
    length_ms = 1000.0 * (double)window->length / window->wave->samp_rate;
    spDebug(50, "swDrawWaveformShading", "length_ms = %f, simplify_thresh = %f, x_factor = %f, x_factor2 = %f\n",
	    length_ms, simplify_thresh, x_factor, x_factor2);
    
    if (swIsWaveProcessing(window->wave) == SP_TRUE
	|| ((window->config->specgram_simplified == SP_TRUE && for_print == SP_FALSE)
	     && length_ms > simplify_thresh)) {
	order_draw_size_min = 2;
	x_update_size = 4;
	ignore_min = window->config->specgram_simplified;
    } else {
	order_draw_size_min = 1;
	x_update_size = 2;
	ignore_min = SP_FALSE;
    }
    spDebug(50, "swDrawWaveformShading", "order_draw_size_min = %d, x_update_size = %d\n",
	    order_draw_size_min, x_update_size);
    
    order_draw_max = /*draw_min + (int)round(order_range * draw_size)*/draw_max;
    /*prev_order = order_min - 1;*/
    prev_order = MAX((long)floor(order_min - 1.0), -1);
    
    for (order = prev_order + 1; (double)order < order_max; order++) {
	if ((double)order + 1.0 >= order_max) {
	    order_draw_min = draw_min;
	} else {
	    /*order_draw_min = draw_min + (int)round((order_max - (double)order - 1.0) * draw_size);*/
	    order_draw_min = draw_min + (int)round((order_max - (double)order) * draw_size);
	}
	order_draw_size = order_draw_max - order_draw_min;

	if (((double)order + 1.0 >= order_max && order_draw_size >= 1)
	    || order_draw_size >= order_draw_size_min) {
	    
	    px = swSampToDrawWidth(window, draw_width_i,
				   swTargetSampToSamp(window, wave, wave_offset + (thin_length * draw_offset)));
	    px += x_offset;
	    
	    flag = SP_FALSE;
	    last_index = -1;
	    last_value = 0.0;
		
	    min = max = 0;
	    min_index = max_index = 0;
	    
	    for (k = 0; k < draw_length; k++) {
		l = draw_offset + k;
		x = (swSampToDrawWidth(window, draw_width_i,
				       swTargetSampToSamp(window, wave, wave_offset + (thin_length * l)))
		     + swSampToDrawWidth(window, draw_width_i,
					 swTargetSampToSamp(window, wave, wave_offset + (thin_length * l) + 1))) / 2;
		x += x_offset;

		for (m = prev_order + 1; m <= order; m++) {
		    if (peak_flag == SP_TRUE) {
			swGetPeakData(wave, channel, m, l * thin_length, &value);
		    } else {
			swGetWaveData(wave, channel, m, l, &value);
		    }

		    if (flag == SP_FALSE) {
			/* initialize max */
			min = value;
			max = value;
			min_index = max_index = k;
			flag = SP_TRUE;
		    } else {
			if (value < min) {
			    min = value;
			    min_index = k;
			}
			if (value > max) {
			    max = value;
			    max_index = k;
			}
		    }
		}
	    
		if (x <= x_offset || x - px < x_update_size || (thin_length >= 2 && l % 2 != 0)) {
#if 0
		    spDebug(100, "swDrawWaveformShading", "x = %d, px = %d, min_index = %ld, max_index = %ld\n",
			    x, px, min_index, max_index);
#endif
		} else {
		    if (x > px) {
			swFillShadingRectangle(component, window, wave, order_draw_min,
					       order_draw_size, x, px, min, max, min_index, max_index,
					       ignore_min, &last_value, &last_index);
		    }
		    flag = SP_FALSE;
		    px = x;
		}
	    }

	    l = draw_offset + k;
	    x = (swSampToDrawWidth(window, draw_width_i,
				   swTargetSampToSamp(window, wave, wave_offset + (thin_length * l)))
		 + swSampToDrawWidth(window, draw_width_i,
				     swTargetSampToSamp(window, wave, wave_offset + (thin_length * l) + 1))) / 2;
	    x += x_offset;
	    
	    if (x + 10 >= x_offset + draw_width_i || draw_length >= window->length) {
		x = x_offset + draw_width_i;
	    }
	    
	    if (x > px) {
		swFillShadingRectangle(component, window, wave, order_draw_min,
				       order_draw_size, x, px, min, max, min_index, max_index,
				       ignore_min, &last_value, &last_index);
	    }
	    
	    order_draw_max = order_draw_min;
	    prev_order = order;
	}
    }
    
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}
#endif

void swDrawWaveformLine(spComponent component, swWave wave, double y_draw_offset,
			double draw_width, double draw_height, double zero_offset,
			int x_offset, double x_factor, double y_factor,
			int channel, spLong offset, spLong length,
			spLong draw_offset, spLong draw_length, spBool peak_flag, spBool for_print)
{
    int x, y;
    int px, py;
    int tx, ty;
    int min, max;
    int pmin, pmax;
    int draw_min, draw_max;
    spLong k, l;
    spLong thin_length;
    double value;
    double zero_height;

    swGetDrawRange(y_draw_offset, draw_height, channel, &draw_min, &draw_max);
    zero_height = (double)draw_min + zero_offset;

    if (peak_flag == SP_TRUE) {
	thin_length = wave->peak_buf_thin_length;
	swGetPeakData(wave, channel, 0, draw_offset * thin_length, &value);
    } else {
	thin_length = wave->thin_length;
	swGetWaveData(wave, channel, 0, draw_offset, &value);
    }
    
    spDebug(50, "swDrawWaveformLine", "x_factor = %f, thin_length = %ld, draw_length = %ld\n",
	    x_factor, thin_length, draw_length);
    
    px = (int)round((double)draw_offset * x_factor) + x_offset;
    py = (int)round(zero_height - (value * y_factor));
    if (x_factor > /*0.2*/0.1 && thin_length < 16) {
	for (k = 1; k < draw_length; k++) {
	    l = draw_offset + k;
	    if (peak_flag == SP_TRUE) {
		swGetPeakData(wave, channel, 0, l * thin_length, &value);
	    } else {
		swGetWaveData(wave, channel, 0, l, &value);
	    }
	    x = (int)round((double)l * x_factor) + x_offset;
	    y = (int)round(zero_height - (value * y_factor));

	    if (y >= draw_min && y <= draw_max) {
		if (py >= draw_min && py <= draw_max) {
		    /* inside --> inside */
		    spDrawLine(component, sw_graphics->gx_fg, px, py, x, y);
		} else {
		    /* outside --> inside */
		    if (py < draw_min) {
			px = px + (int)round((double)(draw_min - py) * (double)(x - px) / (double)(y - py));
			py = draw_min;
		    } else {
			px = px + (int)round((double)(draw_max - py) * (double)(x - px) / (double)(y - py));
			py = draw_max;
		    }
		    
		    spDrawLine(component, sw_graphics->gx_fg, px, py, x, y);
		}
	    } else {
		/* outside of canvas */
		if (py >= draw_min && py <= draw_max) {
		    /* inside --> outside */
		    if (y < draw_min) {
			tx = px + (int)round((double)(draw_min - py) * (double)(x - px) / (double)(y - py));
			ty = draw_min;
		    } else {
			tx = px + (int)round((double)(draw_max - py) * (double)(x - px) / (double)(y - py));
			ty = draw_max;
		    }
		    
		    spDrawLine(component, sw_graphics->gx_fg, px, py, tx, ty);
		} else {
		    /* outside --> outside */
		    if (py < draw_min && y > draw_max) {
			tx = px + (int)round((double)(draw_min - py) * (double)(x - px) / (double)(y - py));
			ty = draw_min;
			px = px + (int)round((double)(draw_max - py) * (double)(x - px) / (double)(y - py));
			py = draw_max;
			spDrawLine(component, sw_graphics->gx_fg, px, py, tx, ty);
		    } else if (py > draw_max && y < draw_min) {
			tx = px + (int)round((double)(draw_max - py) * (double)(x - px) / (double)(y - py));
			ty = draw_max;
			px = px + (int)round((double)(draw_min - py) * (double)(x - px) / (double)(y - py));
			py = draw_min;
			spDrawLine(component, sw_graphics->gx_fg, px, py, tx, ty);
		    }
		}
	    }
	    py = y;
	    px = x;
	}
    } else {
	spDebug(50, "swDrawWaveformLine", "before drawing: px = %d, py = %d\n", px, py);
	x = px;	y = py;
	min = max = py;
	pmin = pmax = py;
	for (k = 1; k < draw_length; k++) {
	    l = draw_offset + k;
	    if (peak_flag == SP_TRUE) {
		swGetPeakData(wave, channel, 0, l * thin_length, &value);
	    } else {
		swGetWaveData(wave, channel, 0, l, &value);
	    }
	    x = (int)round((double)l * x_factor) + x_offset;
	    y = (int)round(zero_height - (value * y_factor));
		
	    if (py < min) min = py;
	    if (py > max) max = py;
	    
	    if (x == px || l % 2 != 0) {
	    } else {
		drawCurrentLine(component, sw_graphics->gx_fg, thin_length, draw_min, draw_max,
				pmin, pmax, min, max, px, py, x, y);
		
		pmin = min; pmax = max;
		
		px = x;
		min = max = y;
	    }
	    py = y;
	}
	spDebug(50, "swDrawWaveformLine", "pmin = %d, pmax = %d, min = %d, max = %d\n", pmin, pmax, min, max);
	
	if (x > px) {
	    drawCurrentLine(component, sw_graphics->gx_fg, thin_length, draw_min, draw_max,
			    pmin, pmax, min, max, px, py, x, y);
	}
	spDebug(50, "swDrawWaveformLine", "drawing done: x = %d, px = %d\n", x, px);
    }

    spDebug(80, "swDrawWaveformLine", "done\n");
    
    return;
}

void swDrawWaveform(spComponent component, swWindow window, swWave wave,
		    double y_draw_offset, double draw_width, double draw_height, double zero_offset,
		    int x_offset, double x_factor, double y_factor,
		    int channel, spLong offset, spLong length,
		    spLong draw_offset, spLong draw_length, spBool peak_flag, spBool for_print)
{
    spDebug(50, "swDrawWaveform", "draw_offset = %ld, draw_length = %ld\n",
	    draw_offset, draw_length);

    if (wave->num_order > 1) {
	swDrawWaveformShading(component, window, wave, y_draw_offset,
			      draw_width, draw_height, x_offset, x_factor, y_factor,
			      channel, offset, length, draw_offset, draw_length, peak_flag, for_print);
    } else {
	swDrawWaveformLine(component, wave, y_draw_offset,
			   draw_width, draw_height, zero_offset, x_offset, x_factor, y_factor,
			   channel, offset, length, draw_offset, draw_length, peak_flag, for_print);
    }

    return;
}

void swFillRegion(spComponent component, spGraphics graphics, swWindow window,
		  double y_draw_offset, double draw_width, double draw_height, int channel, int start_d, int end_d)
{
    int st, ed;
    int width;
    int window_width;

    if (window == NULL || (start_d < 0 && end_d < 0) || start_d == end_d)
	return;

    window_width = (int)draw_width;
    if (start_d < end_d) {
	st = MAX(start_d, 0);
	ed = MIN(end_d, window_width);
    } else {
	st = MAX(end_d, 0);
	ed = MIN(start_d, window_width);
    }

    width = ed - st;

    if (width > 0) {
	spDebug(50, "swFillRegion", "st = %d, width = %d\n", st, width);

	if (channel >= 0) {
	    spFillRectangle(component, graphics, st,
			    (int)spRound((double)channel * draw_height + y_draw_offset),
			    width, (int)spRound(draw_height));
	} else {
#if 0
	    if (draw_height <= 0.0) {
		spFillRectangle(component, graphics, st, (int)spRound(y_draw_offset), width, window->height);
	    } else {
		spFillRectangle(component, graphics, st, (int)spRound(y_draw_offset), width, (int)draw_height);
	    }
#else
	    spFillRectangle(component, graphics, st, (int)spRound(y_draw_offset), width,
			    (int)spRound(draw_height * (double)window->wave->num_channel));
#endif
	}
    }

    return;
}

void swDrawRegionLabels(spComponent component, swWindow window)
{
    long k;
    int st_d, ed_d;

    if (swIsNoLabel(window) == SP_TRUE || window->draw_label == SP_FALSE)
	return;

#if 0
    if (window->drag_label_type == SW_DRAG_NO_LABEL) {
	swSortLabels(window->wave->labels, &window->active_label_index);
	spDebug(30, "swDrawRegionLabels", "sorted active label %ld\n", window->active_label_index);
    }
#endif
    
    spDebug(30, "swDrawRegionLabels", "num_buffer = %ld\n", window->wave->labels->num_buffer);

    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if (window->wave->labels->label[k].time >= 0.0
	    && window->wave->labels->label[k].end_time >= 0.0) {
#if 0
	    long st, ed;
	    st = swDimToSamp(window, window->wave->labels->label[k].time);
	    ed = swDimToSamp(window, window->wave->labels->label[k].end_time);
	    st_d = swSampToDisp(window, st);
	    ed_d = swSampToDisp(window, ed);
#else
	    st_d = swDimToDisp(window, window->wave->labels->label[k].time);
	    ed_d = swDimToDisp(window, window->wave->labels->label[k].end_time);
#endif

	    swFillRegion(component, sw_graphics->gx_region_bg, window,
			 0.0, (double)swGetDrawWidth(window), /*window->draw_height*/window->height, -1, st_d, ed_d);
	}
    }
    
    spDebug(80, "swDrawRegionLabels", "done\n");
    
    return;
}

void swFillBackground(spComponent component, int width, int height)
{
    spFillRectangle(component, sw_graphics->gx_bg, 0, 0, width, height);
    spDebug(80, "swFillBackground", "done\n");
    return;
}

void swDrawBackground(spComponent component, swWindow window)
{
    if (window == NULL)
	return;

    swFillBackground(component, window->width, window->height);
    spDebug(40, "swDrawBackground", "fill rectangle done\n");

#ifndef SW_DRAW_LABEL_CANVAS
    swDrawRegionLabels(component, window);
#endif

    return;
}

swWave swGetTargetWave(swWindow window)
{
    swWave wave;
    
    if (window == NULL) return NULL;

#if 0
    if (swIsSpectrogramVisible(window) == SP_TRUE) {
	wave = window->specgram;
	spDebug(100, "swGetTargetWave", "wave is spectrogram\n");
    } else {
	wave = window->wave;
	spDebug(100, "swGetTargetWave", "wave is waveform\n");
    }
#else
    if (window->first_sub_area == NULL || window->first_sub_area->wave == NULL) {
	spDebug(100, "swGetTargetWave", "not found\n");
	return NULL;
    }
    
    wave = window->first_sub_area->wave;
    spDebug(100, "swGetTargetWave", "wave = %ld\n", (long)wave);
#endif

    return wave;
}

void swDrawWaveImageCore(spComponent component, swWindow window, swWave wave, double y_draw_offset, 
			 double draw_width, double draw_height, double data_min, double data_height,
			 spLong offset, spLong length, spLong draw_offset, spLong draw_pos, 
			 spBool specgram_flag, spBool for_print)
{
    int n;
    int left_margin;
    double x_factor, y_factor;
    double zero_offset;
    
    x_factor = draw_width / (double)(MAX(wave->length - 1, 1));
	
    spDebug(30, "swDrawWaveImageCore", "width = %f, height = %f\n", draw_width, draw_height);
    spDebug(30, "swDrawWaveImageCore", "data_height = %f, x_factor = %f\n", data_height, x_factor);
    spDebug(80, "swDrawWaveImageCore", "draw_offset = %ld, draw_pos = %ld\n", draw_offset, draw_pos);

    if (for_print == SP_TRUE) {
	left_margin = SW_PRINT_LEFT_MARGIN;
    } else {
	left_margin = 0;
    }
    
    for (n = 0; n < wave->num_channel; n++) {
	spDebug(80, "swDrawWaveImageCore", "n = %d / %d\n", n, wave->num_channel);
	swDrawVScale(component, window, wave, y_draw_offset, draw_width, draw_height,
		     n, data_min, data_height, &y_factor, &zero_offset, SP_FALSE, specgram_flag, for_print);
#if 0
#ifdef SW_SUPPORT_METER
	if (for_print == SP_FALSE) {
	    swDrawMeter(component, window, wave, y_draw_offset, draw_height, (int)draw_width, n, 0.0);
	}
#endif
#endif

	if (draw_pos > draw_offset) {
	    swDrawWaveform(component, window, wave,
			   y_draw_offset, draw_width, draw_height, zero_offset,
			   left_margin, x_factor, y_factor, n, offset, length,
			   draw_offset, draw_pos - draw_offset, SP_FALSE, for_print);
	}
	if (for_print == SP_TRUE || draw_offset == 0) {
	    swDrawHScale(component, window, wave, y_draw_offset, draw_width, draw_height,
			 n, offset, length, for_print);
	    swDrawVScale(component, window, wave, y_draw_offset, draw_width, draw_height,
			 n, data_min, data_height, &y_factor, &zero_offset, SP_TRUE, specgram_flag, for_print);
	    spDebug(30, "swDrawWaveImageCore", "y_factor = %f, zero_offset = %f\n",
		    y_factor, zero_offset);
	}
    }
    
    spDebug(80, "swDrawWaveImageCore", "done\n");
    
    return;
}

#if 1
void swDrawSubAreaBackground(spComponent component, swWindow window, swWaveSubArea sub_area)
{
    if (window == NULL || sub_area == NULL)
	return;

    spDebug(40, "swDrawSubAreaBackground", "y_d = %f, height_d = %f\n",
	    sub_area->y_d, sub_area->height_d);
    spFillRectangle(component, sw_graphics->gx_bg, 0, (int)spRound(sub_area->y_d),
		    window->width, (int)spRound(sub_area->height_d));

    return;
}

spLong swDrawWaveSubAreaImage(spComponent component, swWindow window, swWaveSubArea sub_area, spLong draw_pos,
			      double amp_min, double amp_max)
{
    spLong offset = 0, length = 0;
    spLong draw_offset;
    spLong wave_length;
    spLong update_length;
    double data_min, data_max;
    double data_height;
    spBool online_draw = SP_FALSE;
    swWave wave;

    spDebug(40, "swDrawWaveSubAreaImage", "in: created = %d, draw_pos = %ld\n",
	    spIsCreated(component), draw_pos);
    
    if (component == NULL || window == NULL || sub_area == NULL)
	return 0;

    if (sub_area->wave == NULL) {
	swDrawSubAreaBackground(component, window, sub_area);
	return 0;
    }

    wave = sub_area->wave;
    wave_length = wave->length;
	
    if (draw_pos < 0) {
	draw_offset = 0;
    } else {
	if (draw_pos > 0) online_draw = SP_TRUE;
	    
	spDebug(40, "swDrawWaveSubAreaImage", "sub_area->drawn_pos = %ld, wave_length = %ld, wave->num_order = %ld\n",
		sub_area->drawn_pos, wave_length, wave->num_order);
	draw_offset = MAX(sub_area->drawn_pos, 0);
	if (draw_offset >= wave_length) {
	    /* finished case */
	    draw_offset = 0;
	    draw_pos = wave_length;
	} else {
	    if (wave->num_order <= 0 && swIsWaveRangeAvailable(wave) == SP_FALSE) {
		draw_offset = 0;
	    } else if (draw_pos <= 0){
		draw_pos = wave_length/* - draw_offset*/;
	    }
	}
    }
	
    spDebug(40, "swDrawWaveSubAreaImage", "draw_offset = %ld, draw_pos = %ld, amp_min = %f, amp_max = %f\n",
	    draw_offset, draw_pos, amp_min, amp_max);

    if (draw_offset == 0) {
	swDrawSubAreaBackground(component, window, sub_area);

	offset = window->offset;
	length = window->length;
	spDebug(40, "swDrawWaveSubAreaImage", "offset = %ld, length = %ld\n", offset, length);
	
	/*window->draw_width = (double)swGetDrawWidth(window);*/
    }

    if (online_draw == SP_TRUE && swIsWaveSubAreaSpectrogram(sub_area) == SP_FALSE) {
	data_min = -swGetLimitValue(wave->samp_bit);
	data_max = swGetLimitValue(wave->samp_bit);
	data_height = data_max - data_min;
    } else {
	if (amp_max <= amp_min) {
	    data_height = swGetDataHeight(window, wave, &data_min, &data_max);
	} else {
	    data_min = amp_min;
	    data_max = amp_max;
	    data_height = data_max - data_min;
	}
    }
	
    spDebug(40, "swDrawWaveSubAreaImage", "data_min = %f, data_max = %f\n", data_min, data_max);
    spDebug(40, "swDrawWaveSubAreaImage", "y_d = %f, draw_width = %f, draw_height = %f\n",
	    sub_area->y_d, window->draw_width, sub_area->draw_height);
	
    swDrawWaveImageCore(component, window, wave, sub_area->y_d, window->draw_width, sub_area->draw_height,
			data_min, data_height, offset, length, draw_offset, draw_pos,
			sub_area->specgram_flag, SP_FALSE);
    update_length = draw_pos - draw_offset;
	
    if (update_length > 0) {
	sub_area->drawn_pos = draw_pos;
	/*window->drawn_pos = draw_pos;*/
    }
    
    spDebug(30, "swDrawWaveSubAreaImage", "done: update_length = %ld\n", update_length);
    
    return update_length;
}

void swDrawWholeWaveImage(spComponent component, swWindow window)
{
    swWaveSubArea sub_area;
    double amp_min, amp_max;
    long update_length;
    long ret;

    swFillBackground(component, window->width, window->height);
    spDebug(80, "swDrawWholeWaveImage", "swFillBackground done\n");

    if (window->wave != NULL) {
	if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	    /* unselect overview region */
	    spDebug(80, "swDrawWholeWaveImage", "call swReverseOverviewRegion\n");
	    swReverseOverviewRegion(window, window->wave->selected_channel, 
				    swSampToOverviewDisp(window, window->sel_st),
				    swSampToOverviewDisp(window, window->sel_ed));
	    
	}
    }
    
    update_length = 0;
    sub_area = swGetNextWaveSubArea(window, NULL);
    
    while (sub_area != NULL) {
	if (swIsWaveSubAreaVisible(sub_area) == SP_TRUE) {
	    if (sub_area->wave == window->wave) {
		amp_min = window->amp_min;
		amp_max = window->amp_max;
	    } else {
		amp_min = 0.0;
		amp_max = -1.0;
	    }
	    
	    ret = swDrawWaveSubAreaImage(component, window, sub_area, 0, amp_min, amp_max);
	    if (ret > update_length) {
		update_length = ret;
	    }
	}
	sub_area = swGetNextWaveSubArea(window, sub_area);
    }

#ifndef SW_DRAW_LABEL_CANVAS
    swDrawLabels(component, window);
#ifdef SW_SUPPORT_MORPHING
    swDrawAnchors(component, window);
#endif
#endif

    if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	/* select region */
	swRedrawRegion(component, window, SP_TRUE, SP_TRUE);
    }
    
    spDebug(80, "swDrawWholeWaveImage", "done\n");
    
    return;
}
#endif

void swDrawWaveImageCB(spComponent component, swWindow window)
{
    swWave wave;
    swWaveSubArea sub_area;
    spLong drawn_pos;
    
    if (window == NULL) return;

    spDebug(30, "swDrawWaveImageCB", "%s: in\n", window->name);
    
    wave = swGetTargetWave(window);
    sub_area = swGetWaveSubArea(window, wave);
    
    if (window->image == NULL) {
	spDebug(30, "swDrawWaveImageCB", "initial callback\n");
	window->image = component;
	if (wave != NULL) {
	    /* drawing will be performed in the callback */
	    swReadTotalWave(wave, SP_FALSE);
	    spDebug(30, "swDrawWaveImageCB", "swReadTotalWave done\n");

	    if (window->offset > 0 || swSampToTargetSamp(window, wave, window->length) != wave->total_length) {
		spDebug(30, "swDrawWaveImageCB", "window->offset = %ld, window->length = %ld\n",
			window->offset, window->length);
		/* reload wave to display different region */
		swReadWave(wave, SP_FALSE, swSampToTargetSamp(window, wave, window->offset),
			   swSampToTargetSamp(window, wave, window->length));
		if (sub_area != NULL) {
		    sub_area->drawn_pos = 0;
		}
		/*window->drawn_pos = 0;*/
	    }

	    swDrawWholeWaveImage(window->image, window);
	    spRefreshCanvas(window->overview_canvas);
	} else {
	    swDrawBackground(window->image, window);
	}
    } else if (wave == NULL
	       || swIsWaveRangeAvailable(wave) == SP_TRUE) {
	drawn_pos = 0;
	if (sub_area != NULL) {
	    drawn_pos = sub_area->drawn_pos;
	}
	
	spDebug(30, "swDrawWaveImageCB", "drawn_pos = %ld\n", drawn_pos);
	if (drawn_pos <= 0
	    || (wave != NULL &&	drawn_pos >= wave->length)) { /* if drawing has been completed */
	    if (sub_area != NULL) {
		sub_area->drawn_pos = 0;
	    }
	    swDrawWholeWaveImage(component, window);
	}
    }

    spDebug(30, "swDrawWaveImageCB", "%s: done\n", window->name);
    
    return;
}

#ifdef SW_SUPPORT_PRINT
void swPrintWaveImagePage(spComponent component, swWindow window)
{
    spLong offset = 0, length = 0;
    double data_min, data_max;
    double data_height;
    double draw_width, draw_height;
    swWave wave;
    
    wave = swGetTargetWave(window);

    if (window->amp_max <= window->amp_min) {
        data_height = swGetDataHeight(window, wave, &data_min, &data_max);
    } else {
        data_min = window->amp_min;
        data_max = window->amp_max;
        data_height = data_max - data_min;
    }

    offset = window->offset;
    length = window->length;
    spDebug(40, "swPrintWaveImagePage", "offset = %ld, length = %ld\n", offset, length);

    draw_width = (double)window->config->print_width;
    draw_height = (double)(window->config->print_height / window->wave->num_channel);
    draw_height = MAX(draw_height, 2);
    draw_height = MAX(draw_height, SW_MIN_DRAW_HEIGHT);
    spDebug(40, "swPrintWaveImagePage", "draw_width = %f, draw_height = %f\n", draw_width, draw_height);

    swDrawWaveImageCore(component, window, wave, SW_PRINT_TOP_MARGIN,
                        /*window->draw_width, window->draw_height,*/
                        draw_width, draw_height,
                        data_min, data_height, offset, length, 0, wave->length,
                        spIsTrue(wave == window->specgram), SP_TRUE);
	
    return;
}

void swPluginCanvasCB(spComponent component, void *data)
{
    swWindow window;

    if ((window = spGetUserData(component)) != NULL) {
        swPrintWaveImagePage(component, window);
    }
    
    return;
}

spBool swPrintWaveImage(spComponent component, swWindow window)
{
    int width, height;
    double print_factor, print_x_factor, print_y_factor;
    
    if (component == NULL || window == NULL || window->wave == NULL || window->wave->length <= 1) {
        spDebug(40, "swPrintWaveImage", "no data\n");
	return SP_FALSE;
    }

    spSetUserData(component, window);

#if 0
    window->config->print_width = 400;
    window->config->print_height = 350;
#endif

    spSetScalingFactor(component, 1.0, 1.0);
    
    if (spGetClientSize(component, &width, &height) == SP_FALSE) {
        spDebug(40, "swPrintWaveImage", "spGetClientSize failed\n");
        return SP_FALSE;
    }
    
    spDebug(40, "swPrintWaveImage", "width = %d, height = %d\n", width, height);
    
    print_x_factor = (double)width / (double)(window->config->print_width + SW_PRINT_LEFT_MARGIN + SW_PRINT_RIGHT_MARGIN);
    print_y_factor = (double)height / (double)(window->config->print_height + SW_PRINT_TOP_MARGIN + SW_PRINT_BOTTOM_MARGIN);
    print_factor = MIN(print_x_factor, print_y_factor);
    spDebug(30, "swPrintWaveImage", "print_x_factor = %f, print_y_factor = %f\n",
	    print_x_factor, print_y_factor);
    
    spSetScalingFactor(component, print_factor, print_factor);

    if (spStartDrawing(component, window->wave->core->orig_filename) == SP_TRUE) {
	spEndDrawing(component);
        spDebug(40, "swPrintWaveImage", "done\n");
	return SP_TRUE;
    } else {
        spDebug(40, "swPrintWaveImage", "spStartDrawing failed\n");
	return SP_FALSE;
    }
}
#endif

void swCopyImage(swWindow window)
{
    if (window == NULL) return;
    
    spDebug(100, "swCopyImage", "in\n");
    
    /* copy wave image */
    spCopyImage(window->image, window->canvas, 0, 0,
		window->width, window->height, 0, 0);
    
    spDebug(100, "swCopyImage", "done\n");
    
    return;
}

spBool swGetDrawTimeString(swWindow window, double point_f, char *string, spBool align_left)
{
    spLong point;
    char *format1;
    char format[SP_MAX_MESSAGE];
    static char format_left[] = "%-9";
    static char format_right[] = "%9";
    
    if (window == NULL || swIsNoWave(window) == SP_TRUE) return SP_FALSE;
    
    if (align_left == SP_TRUE) {
	format1 = format_left;
    } else {
	format1 = format_right;
    }
    
    if (window->data_type == SW_FREQ_DATA) {
	if (window->config->freq_format == SW_FREQ_FORMAT_KHZ) {
	    sprintf(format, "%s.6f", format1);
	    sprintf(string, format, point_f / 1000.0);
	} else if (window->config->freq_format == SW_FREQ_FORMAT_POINT) {
	    point = swDimToSamp(window, point_f);
	    sprintf(format, "%sld", format1);
	    sprintf(string, format, point);
	} else {
	    sprintf(format, "%s.3f", format1);
	    sprintf(string, format, point_f);
	}
    } else {
	if (window->config->time_format == SW_TIME_FORMAT_POINT) {
	    point = swDimToSamp(window, point_f);
	    sprintf(format, "%sld", format1);
	    sprintf(string, format, point);
	} else {
	    char buf[SP_MAX_MESSAGE];
	    
	    swGetTimeString(point_f, window->config->time_format, buf);
	    sprintf(format, "%ss", format1);
	    sprintf(string, format, buf);
	}
    }

    return SP_TRUE;
}

void swGetTimeStringTitle(swConfig config, swDataType data_type, char *title_string, char *dim_string)
{
    if (data_type == SW_FREQ_DATA) {
	strcpy(title_string, "Frequency");
	if (config->freq_format == SW_FREQ_FORMAT_KHZ) {
	    strcpy(dim_string, "[kHz]");
	} else if (config->freq_format == SW_FREQ_FORMAT_POINT) {
	    strcpy(dim_string, "[points]");
	} else {
	    strcpy(dim_string, "[Hz]");
	}
    } else {
	strcpy(title_string, "Time");
	if (config->time_format == SW_TIME_FORMAT_MSEC
	    || config->time_format == SW_TIME_FORMAT_FLOORED_MSEC) {
	    strcpy(dim_string, "[ms]");
	} else if (config->time_format == SW_TIME_FORMAT_POINT) {
	    strcpy(dim_string, "[points]");
	} else {
	    strcpy(dim_string, "[s]");
	}
    }

    return;
}

void swDrawWaveSubAreaString(swWindow window, swWaveSubArea sub_area)
{
    spBool zero_flag = SP_FALSE;
    int n;
    int sx, sy, swidth, sheight;
    int left_offset;
    spLong point;
    double top_offset;
    double value;
    double point_f;
    double limit, dbvalue;
    spLong data_offset;
    char string[SP_MAX_MESSAGE];
    char label[SP_MAX_MESSAGE];
    char buf[SP_MAX_MESSAGE];
    char *name_str;
    char *unit_str;
    swWave wave;
    
    if (window == NULL || sub_area == NULL || sub_area->wave == NULL) return;

    wave = sub_area->wave;

    top_offset = SW_TIME_STRING_TOP_OFFSET;
    left_offset = SW_TIME_STRING_LEFT_OFFSET;
    if (top_offset < 0) {
	top_offset += /*window->height*/sub_area->height_d;
    }
    if (left_offset < 0) {
	/*left_offset += window->width;*/
	left_offset += (int)window->draw_width;
    }
    if (window->config->scale_flag == SP_TRUE) {
	top_offset = MIN(top_offset, /*window->height*/sub_area->height_d - SW_HSCALE_BOTTOM_SPACING);
	left_offset = MAX(left_offset, SW_HSCALE_LEFT_SPACING);
    }
    top_offset += sub_area->y_d;

    /*point = window->point;*/
    point = swSampToTargetSamp(window, wave, window->point);
    
    if (point < 0 || point >= wave->total_length) {
	point_f = 0.0;
	zero_flag = SP_TRUE;
    } else {
	point_f = window->point_f;
    }

    {
	char title_string[SP_MAX_LINE];
	char time_string[SP_MAX_LINE];
	char dim_string[SP_MAX_LINE];

	swGetDrawTimeString(window, point_f, time_string, SP_FALSE);
	swGetTimeStringTitle(window->config, window->data_type, title_string, dim_string);
	
	sprintf(string, "%s: %s %s", title_string, time_string, dim_string);
	spDebug(100, "swDrawString", "string = %s\n", string);

	if (wave->num_order > 1) {
	    if (swIsWaveSubAreaSpectrogram(sub_area) == SP_TRUE) {
		sprintf(buf, "    Frequency: %9.2f [Hz]",
			(window->wave->samp_rate / 2.0) * (double)window->target_order
			/ (double)wave->num_order);
	    } else {
		sprintf(buf, "    Order: %4ld", window->target_order);
	    }
	    strcat(string, buf);
	}
	
	/* draw time string */
	spDrawString(window->canvas, sw_graphics->gx_string,
		     left_offset, (int)top_offset, string);

	if (swGetTargetWave(window) == wave) {
	    if (window->config->draw_selection_times == SP_TRUE) {
		char end_time_string[SP_MAX_LINE];

		if (window->sel_st >= 0 && window->sel_ed >= 0) {
		    double st_f, ed_f;

		    st_f = swSampToDim(window, window->sel_st);
		    ed_f = swSampToDim(window, window->sel_ed);
		
		    swGetDrawTimeString(window, st_f, time_string, SP_FALSE);
		    swGetDrawTimeString(window, ed_f, end_time_string, SP_TRUE);
		} else {
		    swGetDrawTimeString(window, 0.0, time_string, SP_FALSE);
		    swGetDrawTimeString(window, 0.0, end_time_string, SP_TRUE);
		}
	    
		sprintf(string, "Selection: %s - %s %s", time_string, end_time_string, dim_string);
	    
		if (spGetStringExtent(window->canvas, sw_graphics->gx_string, string, &sx, &sy, &swidth, &sheight, NULL) == SP_TRUE) {
		    top_offset -= (-sy + 2);
		} else {
		    top_offset -= SW_TIME_STRING_HEIGHT;
		}

		/* draw selection string */
		spDrawString(window->canvas, sw_graphics->gx_string,
			     left_offset, (int)top_offset, string);
	    }

	    if (window->config->draw_selection_length == SP_TRUE) {
		spLong len = 0;
		double len_f = 0.0;

		if (window->sel_st >= 0 && window->sel_ed >= 0) {
		    if (window->sel_st < window->sel_ed) {
			len = window->sel_ed - window->sel_st;
		    } else {
			len = window->sel_st - window->sel_ed;
		    }
		    len_f = swSampToDim(window, len);
		
		    swGetDrawTimeString(window, len_f, time_string, SP_FALSE);
		} else {
		    swGetDrawTimeString(window, 0.0, time_string, SP_FALSE);
		}
	    
		if (window->data_type == SW_FREQ_DATA) {
		    sprintf(string, "Length: %s %s", time_string, dim_string);
		} else {
		    char freq_string[SP_MAX_LINE];

		    if (len_f > 0.0) {
			sprintf(freq_string, "%.4f", 1.0/len_f);
		    } else {
			sprintf(freq_string, "oo");
		    }
		    sprintf(string, "Length: %s %s (%9s [Hz])", time_string, dim_string, freq_string);
		}
	    
		if (spGetStringExtent(window->canvas, sw_graphics->gx_string, string, &sx, &sy, &swidth, &sheight, NULL) == SP_TRUE) {
		    top_offset -= (-sy + 2);
		} else {
		    top_offset -= SW_TIME_STRING_HEIGHT;
		}

		/* draw selection string */
		spDrawString(window->canvas, sw_graphics->gx_string,
			     left_offset, (int)top_offset, string);
	    }
	}
    }

#ifdef SW_SUPPORT_METER
    if (swGetTargetWave(window) == wave) {
	for (n = 0; n < wave->num_channel; n++) {
	    if (swGetPeakData(wave, n, window->target_order, window->point, &value) == SP_FALSE) {
		value = 0.0;
	    }
	    swDrawMeter(window->canvas, window, wave, /*sub_area->y_d*/0.0,
			/*sub_area->draw_height*/(double)window->height / (double)MAX(wave->num_channel, 1),
			(int)window->draw_width, n, value);
	}
    }
#endif
    
    if (zero_flag == SP_TRUE || swIsWaveProcessing(wave) == SP_TRUE
	/*|| wave->num_order > 1*/) return;

    top_offset = SW_AMP_STRING_TOP_OFFSET;
    left_offset = SW_AMP_STRING_LEFT_OFFSET;
    if (top_offset < 0) {
	top_offset += sub_area->height_d;
    }
    if (left_offset < 0) {
	left_offset += (int)window->draw_width;
    }
    if (window->config->scale_flag == SP_TRUE) {
	top_offset = MIN(top_offset, /*window->height*/sub_area->height_d - SW_HSCALE_BOTTOM_SPACING);
	left_offset = MAX(left_offset, SW_HSCALE_LEFT_SPACING);
    }
    top_offset += sub_area->y_d;

    /*data_offset = swSampToTargetSamp(window, wave, window->point - window->offset) / wave->thin_length;*/
    data_offset = (swSampToTargetSamp(window, wave, window->point) - swGetWaveOffset(wave)) / wave->thin_length;
    
    spDebug(40, "swDrawString", "data_offset = %ld, point = %ld, offset = %ld, target_order = %ld\n",
	    data_offset, window->point, window->offset, window->target_order);
    spDebug(40, "swDrawString", "length = %ld, total_length = %ld, num_channel = %d\n",
	    wave->length, wave->total_length, wave->num_channel);

    limit = swGetWindowLimitValue(window, wave);
    
    for (n = 0; n < wave->num_channel; n++) {
	if (swGetWaveData(wave, n, window->target_order, data_offset, &value) == SP_FALSE) {
	    break;
	}

	name_str = swGetAnalysisNameString(wave);
	if (!strnone(name_str)) {
	    strcpy(label, name_str);
	} else {
	    strcpy(label, "Amplitude");
	}
	
	unit_str = swGetAnalysisUnitString(wave);
	
	if (wave->num_order <= 1
	    && window->config->percent_amplitude == SP_TRUE
	    && window->data_type != SW_FREQ_DATA && strnone(unit_str)) {
	    sprintf(string, "%s: %8.3f %%", label, 100.0 * value / limit);
	} else {
	    if (swIsWaveFloat(wave) == SP_TRUE) {
		sprintf(string, "%s:%10.4f", label, value);
	    } else if (swIsWaveLong(wave) == SP_TRUE) {
		sprintf(string, "%s:%10.0f", label, value);
	    } else {
		sprintf(string, "%s:%7.0f", label, value);
	    }

	    if (!strnone(unit_str)) {
		strcat(string, " ");
		strcat(string, unit_str);
	    }
	}

	if (wave->num_order <= 1
	    && window->data_type != SW_FREQ_DATA && strnone(unit_str)) {
	    dbvalue = FABS(value) / limit;
	    if (dbvalue <= 1e-30) {
		sprintf(buf, " (%8s dB)", "-oo");
	    } else {
		sprintf(buf, " (%8.3f dB)", dB(dbvalue));
	    }
	
	    strcat(string, buf);
	}

	spDebug(50, "swDrawString", "string = %s\n", string);

	/* draw amplitude string */
	spDrawString(window->canvas, sw_graphics->gx_string,
		     left_offset, (int)top_offset, string);

	top_offset += sub_area->draw_height;
    }
    
    return;
}

void swDrawString(swWindow window)
{
    swWaveSubArea sub_area;
    
    sub_area = swGetNextWaveSubArea(window, NULL);
    
    while (sub_area != NULL) {
	if (swIsWaveSubAreaVisible(sub_area) == SP_TRUE) {
	    swDrawWaveSubAreaString(window, sub_area);
	}
	sub_area = swGetNextWaveSubArea(window, sub_area);
    }
    
    return;
}

static int sw_label_string_px = 0;
static int sw_label_string_py = 0;
static int sw_label_string_prev_channel = -1;

static int sw_region_label_px = 0;
static int sw_region_label_height = 0;
static int sw_region_label_prev_channel = -1;

#define SW_LABEL_STRING_OFFSET 5
/*#define SW_LABEL_STRING_TOP_OFFSET 25*/
#define SW_LABEL_STRING_TOP_OFFSET 10
#define SW_REGION_LABEL_OFFSET 8
#define SW_REGION_LABEL_BOTTOM_OFFSET 23

#define SW_ACTIVE_LABEL_RECT_SIZE 6

void swDrawChannelLabel(spComponent component, int draw_width, int window_height, int num_channel,
			int st_d, int ed_d, int channel, char *string, spBool active_flag)
{
    int cx, cy;
    int x, y;
    int width, height;
    int y_offset;
    int label_height;
    double region_height;
    spGraphics graphics;
    
    spLockMutex(sw_graphics->mutex);

    if (channel < 0) {
	y_offset = 0;
	label_height = window_height;
    } else {
	y_offset = (int)spRound((double)channel * (double)window_height / (double)num_channel);
	label_height = (int)spRound((double)window_height / (double)num_channel);
    }
    
    if (ed_d >= 0) {
	graphics = sw_graphics->gx_region;

	spSetGraphicsParams(graphics,
			    SppForeground, sw_graphics->region_line_color,
			    NULL);

	/* draw arrow */
	if (st_d >= 0 && sw_region_label_px > st_d && sw_region_label_prev_channel == channel) {
	    region_height = sw_region_label_height + SW_REGION_LABEL_OFFSET;
	    if (region_height > label_height) {
		region_height = SW_REGION_LABEL_BOTTOM_OFFSET;
	    }
	} else {
	    region_height = SW_REGION_LABEL_BOTTOM_OFFSET;
	}
	    
	y = y_offset + label_height - (int)region_height;
	
	if (ed_d <= draw_width) {
	    if (channel >= 0) {
		spSetGraphicsParams(graphics, SppLineType, SP_LINE_DASH, NULL);
		spDrawLine(component, graphics, ed_d, 0, ed_d, window_height);
		spSetGraphicsParams(graphics, SppLineType, SP_LINE_SOLID, NULL);
	    }
	    spDrawLine(component, graphics, ed_d, y_offset, ed_d, y_offset + label_height);
	    
	    if (active_flag == SP_TRUE) {
		spFillRectangle(component, graphics, ed_d - SW_ACTIVE_LABEL_RECT_SIZE/2, y_offset,
				SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
		spFillRectangle(component, graphics, ed_d - SW_ACTIVE_LABEL_RECT_SIZE/2,
				y_offset + label_height - SW_ACTIVE_LABEL_RECT_SIZE,
				SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
	    }
	    
	    spSetGraphicsParams(graphics,
				SppForeground, sw_graphics->region_label_color,
				NULL);

	    /* arrow's head */
	    spDrawLine(component, graphics, ed_d, y, ed_d - 9, y - 3);
	    spDrawLine(component, graphics, ed_d, y, ed_d - 9, y + 3);
	    spDrawLine(component, graphics, MAX(st_d, 0), y, ed_d, y);

	    sw_region_label_px = MAX(ed_d, sw_region_label_px);
	    sw_region_label_height = (int)region_height;
	    sw_region_label_prev_channel = channel;
	} else {
	    spSetGraphicsParams(graphics,
				SppForeground, sw_graphics->region_label_color,
				NULL);

	    if (st_d <= draw_width) {
		/* arrow's body */
		spDrawLine(component, graphics, MAX(st_d, 0), y, MIN(ed_d, draw_width), y);
	    }
	}
    } else {
	graphics = sw_graphics->gx_label;
    }
    
    if (st_d >= 0 && st_d <= draw_width) {
	if (channel >= 0) {
	    spSetGraphicsParams(graphics, SppLineType, SP_LINE_DASH, NULL);
	    spDrawLine(component, graphics, st_d, 0, st_d, window_height);
	    spSetGraphicsParams(graphics, SppLineType, SP_LINE_SOLID, NULL);
	}
	spDrawLine(component, graphics, st_d, y_offset, st_d, y_offset + label_height);
	
	if (active_flag == SP_TRUE) {
	    spFillRectangle(component, graphics, st_d - SW_ACTIVE_LABEL_RECT_SIZE/2, y_offset,
			    SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
	    spFillRectangle(component, graphics, st_d - SW_ACTIVE_LABEL_RECT_SIZE/2,
			    y_offset + label_height - SW_ACTIVE_LABEL_RECT_SIZE,
			    SW_ACTIVE_LABEL_RECT_SIZE, SW_ACTIVE_LABEL_RECT_SIZE);
	}
	
	if (!strnone(string)) {
	    spGetStringExtent(component, graphics, string, &x, &y, &width, &height, NULL);
	    
	    cx = st_d;
	    if (sw_label_string_px >= cx && sw_label_string_prev_channel == channel) {
		cy = y_offset + sw_label_string_py + height + SW_LABEL_STRING_OFFSET;
		if (cy > label_height) {
		    cy = y_offset + height + SW_LABEL_STRING_TOP_OFFSET;
		}
	    } else {
		cy = y_offset + height + SW_LABEL_STRING_TOP_OFFSET;
	    }
	    
	    sw_label_string_px = cx + width;
	    sw_label_string_py = cy;
	    sw_label_string_prev_channel = channel;
	    
	    spDrawLine(component, graphics, cx, cy, sw_label_string_px, cy);
	    spDrawString(component, graphics, st_d, cy - 1, string);
	}
    }

    if (ed_d >= 0) {
	spSetGraphicsParams(sw_graphics->gx_region,
			    SppForeground, sw_graphics->region_color,
			    NULL);
    }
    
    spUnlockMutex(sw_graphics->mutex);
    
    return;
}

void swDrawLabel(spComponent component, int draw_width, int window_height,
		 int st_d, int ed_d, char *string, spBool active_flag)
{
    swDrawChannelLabel(component, draw_width, window_height, 1,
		       st_d, ed_d, -1, string, active_flag);
    
    return;
}

void swDrawLabels(spComponent component, swWindow window)
{
    long k;
    int st_d, ed_d;
    spBool active_flag;

    if (swIsNoLabel(window) == SP_TRUE
	|| window->draw_label == SP_FALSE || component == NULL)
	return;

    spDebug(80, "swDrawLabels", "in\n");
    
    sw_label_string_px = 0;
    sw_label_string_py = SW_LABEL_STRING_TOP_OFFSET;
    sw_label_string_prev_channel = -1;
    sw_region_label_px = 0;
    sw_region_label_height = SW_REGION_LABEL_BOTTOM_OFFSET;
    sw_region_label_prev_channel = -1;

    for (k = 0; k < window->wave->labels->num_buffer; k++) {
	if (window->wave->labels->label[k].time >= 0.0) {
	    st_d = swDimToDisp(window, window->wave->labels->label[k].time);

	    if (k == window->active_label_index) {
		active_flag = SP_TRUE;
	    } else {
		active_flag = SP_FALSE;
	    }

	    if (window->wave->labels->label[k].end_time >= 0.0) {
		ed_d = swDimToDisp(window, window->wave->labels->label[k].end_time);
	    } else {
		ed_d = -1;
	    }
	    swDrawChannelLabel(component, (int)window->draw_width, window->height, window->wave->num_channel,
			       st_d, ed_d, window->wave->labels->label[k].channel,
			       window->wave->labels->label[k].string, active_flag);
	}
    }

    spDebug(80, "swDrawLabels", "done\n");
    
    return;
}

void swRedrawLabels(swWindow window)
{
#ifndef SW_DRAW_LABEL_CANVAS
    swRedrawWave(window);
#else
    swDrawCursor(window, SP_TRUE);
#endif
    
    return;
}

void swUpdateLabels(swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE) return;

    if (swIsLabelNone(window->wave) == SP_TRUE) {
	window->active_label_index = -1;
    } else {
	if (window->drag_label_type == SW_DRAG_NO_LABEL) {
	    swSortLabels(window->wave->labels, &window->active_label_index);
	    spDebug(50, "swUpdateLabels", "sorted active label %ld\n", window->active_label_index);
	}
	
	if (window->active_label_index >= 0
	    && window->wave->labels->label[window->active_label_index].time < 0.0) {
	    window->active_label_index = -1;
	}
    }
    
    swRedrawLabels(window);
    swUpdateLabelList(window->label_list);

#if defined(SW_AH_CUSTOM)
    if (window->link_ah_file) {
	swLabelsToAHFile(window->wave, window->config->toplevel->ahfile);
	swDrawAHSessionCanvas(window->config->toplevel->ahwindow, SP_TRUE, SP_TRUE);
    }
#endif
    
    return;
}

void swUnselectActiveLabel(swWindow window)
{
    if (swIsNoWave(window) == SP_TRUE) {
	return;
    }
    
    if (window->active_label_index >= 0) {
	window->active_label_index = -1;
	swRedrawLabels(window);
    }

    return;
}

void swResetRegion(swWindow window)
{
    if (window == NULL) return;

    window->sel_st = -1;
    window->sel_ed = -1;
    window->sel_st_d = -1;
    window->sel_ed_d = -1;
	
    swSetSelectSenseLevel(window, SP_FALSE);
    
    return;
}

void swUpdateDisplayRegion(swWindow window)
{
    if (window->sel_st == -1 || window->sel_ed == -1) {
	window->sel_st_d = -1;
	window->sel_ed_d = -1;
    } else {
	window->sel_st_d = swSampToDisp(window, window->sel_st);
	window->sel_ed_d = swSampToDisp(window, window->sel_ed);
    
	if (window->sel_st != window->sel_ed && 
	    window->sel_st_d == window->sel_ed_d) {
	    window->sel_ed_d += 1;
	}
    }
    
    return;
}

void swSelectRegion(swWindow window, int channel, spLong start, spLong end)
{
    if (window == NULL || window->wave == NULL)	return;

    spDebug(50, "swSelectRegion", "in\n");
    
    swLockWindowMutex(window);

    if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	swReverseRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d,
			SP_TRUE, SP_TRUE);
	spDebug(50, "swSelectRegion", "reverse done\n");
    }

    if ((start < 0 && end < 0) || 
	(start >= window->wave->total_length && end >= window->wave->total_length)) {
	swResetRegion(window);
    } else {
	if (start < end) {
	    window->sel_st = MAX(start, 0);
	    window->sel_ed = MIN(end, window->wave->total_length - 1);
	} else {
	    window->sel_st = MAX(end, 0);
	    window->sel_ed = MIN(start, window->wave->total_length - 1);
	}
	
	if (window->sel_st == window->sel_ed) {
	    swResetRegion(window);
	} else {
	    swUpdateDisplayRegion(window);
	    
	    window->wave->selected_channel = channel;

	    swUpdateInfoAreaSelection(window);
	    
	    /* redraw region */
	    swReverseRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d,
			    SP_TRUE, SP_TRUE);

	    swSetSelectSenseLevel(window, SP_TRUE);
	}
    }

    /* draw cursor */
    swDrawCursor(window, SP_TRUE);
    /*spRefreshCanvas(window->overview_canvas);*/
    
    swUnlockWindowMutex(window);
    
    spDebug(50, "swSelectRegion", "done\n");
    
    return;
}

void swSelectRegionCB(spComponent component, swWindow window)
{
    spLong start, end;

    if (window == NULL || window->wave == NULL) return;
    
    start = window->offset;
    end = start + window->length - 1;
    
    if (start == window->sel_st && end == window->sel_ed) {
	start = -1;
	end = -1;
    }
    
    swSelectRegion(window, -1, start, end);

    return;
}

void swSelectFromHereCB(spComponent component, swWindow window)
{
    int channel;
    spLong start, end;

    if (window == NULL || window->wave == NULL) return;
    
    start = window->point;
    if (window->sel_ed > start) {
	end = window->sel_ed;
	channel = window->wave->selected_channel;
    } else {
	end = window->wave->total_length - 1;
	channel = -1;
    }
    
    swSelectRegion(window, channel, start, end);

    return;
}

void swSelectToHereCB(spComponent component, swWindow window)
{
    int channel;
    spLong start, end;

    if (window == NULL || window->wave == NULL) return;
    
    end = window->point;
    if (window->sel_st >= 0 && window->sel_st < end) {
	start = window->sel_st;
	channel = window->wave->selected_channel;
    } else {
	start = 0;
	channel = -1;
    }
    
    swSelectRegion(window, channel, start, end);

    return;
}

/*
 *	select region of other window
 */
void swSelectAllRegionCB(spComponent component, swWindow window)
{
    spLong k;
    spLong sel_st, sel_ed;
    double sel_st_f, sel_ed_f;
    spComponent next = NULL;
    swWindow next_window = NULL;

    if (window == NULL || window->wave == NULL
	|| (window->sel_st < 0 && window->sel_ed < 0))	return;

    sel_st_f = swSampToDim(window, window->sel_st);
    sel_ed_f = swSampToDim(window, window->sel_ed);

    /* draw cursor */
    swDrawCursor(window, SP_TRUE);

    next = window->window;
    for (k = 0;; k++) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}
	
	if ((next_window = (swWindow)spGetUserData(next)) != NULL
	    && next_window->wave != NULL) {
	    if (next_window->data_type == window->data_type
		&& swIsWaveProcessing(next_window->wave) == SP_FALSE) {
		sel_st = swDimToSamp(next_window, sel_st_f);
		sel_ed = swDimToSamp(next_window, sel_ed_f);

		/* select region */
		swSelectRegion(next_window, -1, sel_st, sel_ed);
	    }
	}
    }

    return;
}

void swSelectAllCB(spComponent component, swWindow window)
{
    spLong start, end;

    if (window == NULL || window->wave == NULL) return;
    
    start = 0;
    end = window->wave->total_length - 1;
    
    swSelectRegion(window, -1, start, end);

    return;
}

void swSelectNextChannelCB(spComponent component, swWindow window)
{
    int channel;

    if (window == NULL || window->wave == NULL) return;

    
    if (window->wave->selected_channel >= window->wave->num_channel - 1) {
	channel = -1;
    } else if (window->wave->selected_channel < 0) {
	channel = 0;
    } else {
	channel = window->wave->selected_channel + 1;
    }
    
    swSelectRegion(window, channel, window->sel_st, window->sel_ed);

    return;
}

void swReverseMainRegion(spComponent component, swWindow window, int channel, int start_d, int end_d)
{
    swWaveSubArea sub_area;
    double draw_width;
    
    if (window == NULL || component == NULL || (start_d < 0 && end_d < 0)) return;

    draw_width = (double)swGetDrawWidth(window);
    spDebug(80, "swReverseMainRegion", "draw_width = %f, start_d = %d, end_d = %d\n",
	    draw_width, start_d, end_d);
    
    sub_area = swGetNextWaveSubArea(window, NULL);
		    
    while (sub_area != NULL) {
	if (swIsWaveSubAreaVisible(sub_area) == SP_TRUE) {
	    swFillRegion(component, sw_graphics->gx_xor, window,
			 sub_area->y_d, draw_width, sub_area->draw_height, channel, start_d, end_d);
	}
	sub_area = swGetNextWaveSubArea(window, sub_area);
    }
    
    return;
}

void swReverseRegionForComponent(spComponent component, swWindow window, int channel,
				 int start_d, int end_d, spBool main_flag, spBool overview_flag)
{
    if (window == NULL) return;

    spDebug(80, "swReverseRegionForComponent", "in: start_d = %d, end_d = %d\n", start_d, end_d);
    
    if (main_flag == SP_TRUE) {
	swReverseMainRegion(component, window, channel, start_d, end_d);
    }

    if (overview_flag == SP_TRUE && window->overview_canvas != NULL) {
	spDebug(80, "swReverseRegionForComponent", "draw overview region\n");
	
	if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	    spLong start, end;
    
	    start = swDispToSamp(window, start_d);
	    end = swDispToSamp(window, end_d);
	    swReverseOverviewRegion(window, channel,
				    swSampToOverviewDisp(window, start),
				    swSampToOverviewDisp(window, end));
	} else {
	    swDrawOverview(window, SP_FALSE);
	}
    }
    
    spDebug(80, "swReverseRegionForComponent", "done\n");
    
    return;
}

void swReverseRegion(swWindow window, int channel,
		     int start_d, int end_d, spBool main_flag, spBool overview_flag)
{
    spComponent component;
    
    if (window == NULL) return;

    if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	component = window->image;
    } else {
	component = window->canvas;
    }
    swReverseRegionForComponent(component, window, channel, start_d, end_d,
				main_flag, overview_flag);
    
    return;
}

void swRedrawRegion(spComponent component, swWindow window, spBool main_flag, spBool overview_flag)
{
    if (window == NULL || (window->sel_st < 0 && window->sel_ed < 0))
	return;
    
    spDebug(30, "swRedrawRegion", "set_st = %ld, sel_ed = %ld\n",
	    (long)window->sel_st, (long)window->sel_ed);
    
    swUpdateDisplayRegion(window);

    spDebug(30, "swRedrawRegion", "set_st_d = %d, sel_ed_d = %d\n",
	    window->sel_st_d, window->sel_ed_d);
    
    swReverseRegionForComponent(component, window, window->wave->selected_channel,
				window->sel_st_d, window->sel_ed_d, main_flag, overview_flag);

    spDebug(80, "swRedrawRegion", "done\n");
    
    return;
}

void swClearRegion(swWindow window)
{
    if (window == NULL) return;
    
    spDebug(30, "swClearRegion", "sel_st_d = %d, sel_ed_d = %d\n",
	    window->sel_st_d, window->sel_ed_d);

    if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	/* clear region */
	swReverseRegion(window, window->wave->selected_channel, window->sel_st_d, window->sel_ed_d,
			SP_TRUE, SP_TRUE);
    }
    
    window->sel_st_d = -1;
    window->sel_ed_d = -1;
    window->sel_st = -1;
    window->sel_ed = -1;

    /* refresh window */
    swDrawCursor(window, SP_TRUE);

    if (window->overview_canvas != NULL) {
	spRefreshCanvas(window->overview_canvas);
    }
    
    return;
}

static void drawHCursor(swWindow window, spComponent canvas, spGraphics graphics, int point_d)
{
    /* draw cursor line */
    spDrawLine(canvas, graphics,
	       point_d, 0, point_d, window->height);
    return;
}

static void drawOrderHCursor(swWindow window, swWaveSubArea sub_area, spComponent canvas, spGraphics graphics, int point_d)
{
    int n;
    int height;
    int offset;
    double draw_min;
    int x_min, x_max;
    swWave wave;

    wave = sub_area->wave;

    if (wave->num_order <= 1
	|| window->target_order < 0 || window->target_order >= wave->num_order) return;
    
    if (window->amp_max <= window->amp_min) {
	offset = (int)round(sub_area->draw_height * (double)(wave->num_order - window->target_order)
			    / (double)wave->num_order);
    } else {
	offset = (int)round(sub_area->draw_height * (window->amp_max - (double)window->target_order)
			    / (window->amp_max - window->amp_min));
    }
    draw_min = sub_area->y_d;

    if (0) {
	int size = 8;
	x_min = point_d - size;
	x_max = point_d + size;
    } else {
	x_min = 0;
	x_max = swGetDrawWidth(window);
    }
	    
    for (n = 0; n < wave->num_channel; n++) {
	height = (int)round(draw_min) + offset;

	/* draw cursor line */
	spDrawLine(canvas, graphics,
		   x_min, height, x_max, height);
	
	draw_min += sub_area->draw_height;
    }
    
    return;
}

void swDrawHCursor(swWindow window)
{
    swWaveSubArea sub_area;
    
    if (window->point_d >= 0 && window->point_d < swGetDrawWidth(window)) {
	drawHCursor(window, window->canvas,
		    sw_graphics->gx_pointer, window->point_d);
	
	sub_area = swGetNextWaveSubArea(window, NULL);
		    
	while (sub_area != NULL) {
	    if (swIsWaveSubAreaVisible(sub_area) == SP_TRUE) {
		drawOrderHCursor(window, sub_area, window->canvas,
				 sw_graphics->gx_pointer, window->point_d);
	    }
	    sub_area = swGetNextWaveSubArea(window, sub_area);
	}
    }
    
    return;
}

void swRefreshWindowWithSelection(swWindow window, int channel, int start_d, int end_d,
				  spBool draw_cursor, spBool overview_flag)
{
    if (window->canvas == NULL) return;
    
    spDebug(100, "swRefreshWindow", "%s: in: draw_cursor = %d, overview_flag = %d\n",
	    window->name, draw_cursor, overview_flag);

    /* copy wave image */
    swCopyImage(window);

    if (window->config->toplevel->rubber_band_selection == SP_FALSE) {
	if (window->wave != NULL && window->overview_canvas != NULL) {
	    swReverseRegionForComponent(window->canvas, window, channel, start_d, end_d,
					SP_TRUE, overview_flag); /* reverse main and overview */
	    if (overview_flag == SP_TRUE) {
		spRefreshCanvas(window->overview_canvas);	   /* refresh overview canvas only (refresh main canvas later) */
	    }
	}
    }

#ifdef SW_DRAW_LABEL_CANVAS
#ifdef SW_SUPPORT_MORPHING
    swDrawAnchors(window->canvas, window);
#endif
    swDrawRegionLabels(window->canvas, window);
    swDrawLabels(window->canvas, window);
#endif

    if (draw_cursor == SP_TRUE) {
	/* draw horizontal cursor */
	swDrawHCursor(window);

	/* draw string */
	swDrawString(window);
    }
    
    spDebug(100, "swRefreshWindow", "%s: before spRefreshCanvas\n", window->name);
    
    /* refresh window */
    spRefreshCanvas(window->canvas);

    spDebug(100, "swRefreshWindow", "%s: done\n", window->name);
    
    return;
}

void swRefreshWindow(swWindow window, spBool draw_cursor, spBool overview_flag)
{
    int channel;

    if (window == NULL) return;

    if (window->wave != NULL) {
	channel = window->wave->selected_channel;
    } else {
	channel = -1;
    }
    
    swRefreshWindowWithSelection(window, channel, window->sel_st_d, window->sel_ed_d,
				 draw_cursor, overview_flag);
    
    return;
}

void swDrawCursorWithSelection(swWindow window, int channel, int start_d, int end_d, spBool overview_flag)
{
    swRefreshWindowWithSelection(window, channel, start_d, end_d,
				 spIsFalse(swIsProcessing(window)), overview_flag);
    
    return;
}

void swDrawCursor(swWindow window, spBool overview_flag)
{
    swRefreshWindow(window, spIsFalse(swIsProcessing(window)), overview_flag);
    
    return;
}


static spBool swGetOverviewEdgeDiff(swWindow window, int x, int *stdiff, int *eddiff,
				    spBool *thumb_flag)
{
    int st_d, ed_d;
    
    if (window == NULL || window->wave == NULL
	|| (swIsWaveProcessing(window->wave) == SP_TRUE
	    && (swIsWavePlaying(window->wave) == SP_FALSE
		/*|| window->sync_play == SP_TRUE*/
		/*|| swIsWaveThreadSafe(window->wave) == SP_FALSE*/))) {
	return SP_FALSE;
    }

    st_d = swSampToOverviewDisp(window, window->offset);
    ed_d = swSampToOverviewDisp(window, window->offset + window->length - 1);

    *stdiff = x - st_d;
    *eddiff = x - ed_d;

    if (ed_d - st_d <= SW_EDGE_MOTION_RANGE * 2
	&& (x > st_d && x < ed_d)) {
	*thumb_flag = SP_TRUE;
    } else {
	*thumb_flag = SP_FALSE;
    }

    return SP_TRUE;
}

static swOverviewDragType swGetOverviewDragType(swWindow window,
						int stdiff, int eddiff, spBool thumb_flag)
{
    swOverviewDragType drag_type = SW_OVERVIEW_DRAG_NONE;

    if (thumb_flag == SP_FALSE && ABS(stdiff) <= SW_EDGE_MOTION_RANGE / 2
	&& ABS(stdiff) < ABS(eddiff)) {
	drag_type = SW_OVERVIEW_DRAG_START;
    } else if (thumb_flag == SP_FALSE && ABS(eddiff) <= SW_EDGE_MOTION_RANGE / 2) {
	drag_type = SW_OVERVIEW_DRAG_END;
    } else if (stdiff >= 0 && eddiff <= 0) {
	drag_type = SW_OVERVIEW_DRAG_THUMB;
    }
	
    return drag_type;
}

void swOverviewPointerMotionCB(spComponent component, swWindow window)
{
    int x, y;
    int stdiff, eddiff;
    spBool thumb_flag;
    swOverviewDragType drag_type = SW_OVERVIEW_DRAG_NONE;
    static swOverviewDragType prev_drag_type = SW_OVERVIEW_DRAG_NONE;
    
    if (swIsNoWave(window) == SP_TRUE || window->overview_canvas == NULL) return;

    if (spGetCallbackMousePosition(component, &x, &y) == SP_TRUE) {
	spDebug(100, "swOverviewPointerMotionCB",
		"component = %ld, overview_canvas = %ld, canvs = %ld\n",
		component, window->overview_canvas, window->canvas);
	x = MAX(x, 0);
	x = MIN(x, window->overview_width);
	if (swGetOverviewEdgeDiff(window, x, &stdiff, &eddiff, &thumb_flag) == SP_TRUE) {
	    drag_type = swGetOverviewDragType(window, stdiff, eddiff, thumb_flag);
	    if (drag_type == SW_OVERVIEW_DRAG_START) {
		swSetCanvasCursor(component, SP_CURSOR_W_RESIZE);
	    } else if (drag_type == SW_OVERVIEW_DRAG_END) {
		swSetCanvasCursor(component, SP_CURSOR_E_RESIZE);
	    } else if (drag_type == SW_OVERVIEW_DRAG_THUMB) {
		swSetCanvasCursor(component, SP_CURSOR_MOVE);
	    } else {
		drag_type = SW_OVERVIEW_DRAG_NONE;
	    }
	}
    }

    if (drag_type == SW_OVERVIEW_DRAG_NONE && prev_drag_type != SW_OVERVIEW_DRAG_NONE) {
	spUnsetCanvasCursor(component);
    }
    prev_drag_type = drag_type;
    
    return;
}

void swOverviewButtonCB(spComponent component, swWindow window)
{
    int x, y;
    int stdiff, eddiff;
    spLong offset, length;
    spBool thumb_flag;
    spCallbackReason reason;
    spModifierMask modmask;
    static spBool overview_play_start = SP_FALSE;
    static swOverviewDragType drag_type = SW_OVERVIEW_DRAG_NONE;
    static spLong offset_diff = 0;
    static spLong overview_play_start_offset = -1;
    
    if (swIsNoWave(window) == SP_TRUE || window->overview_canvas == NULL) return;

    spDebug(50, "swOverviewButton", "in\n");
    
    if (spGetCallbackMousePosition(component, &x, &y) == SP_TRUE) {
	x = MAX(x, 0);
	x = MIN(x, window->overview_width);
	
	reason = spGetCallbackReason(component);
	spDebug(50, "swOverviewButton", "reason = %d\n", reason);
	
	if (swGetOverviewEdgeDiff(window, x, &stdiff, &eddiff, &thumb_flag) == SP_TRUE) {
	    switch (reason) {
	      case SP_CR_LBUTTON_PRESS:
		modmask = 0;
		spGetModifierKeyMask(window->window, &modmask);
		
		drag_type = swGetOverviewDragType(window, stdiff, eddiff, thumb_flag);
		if (modmask & SP_COMMAND_MODIFIER_MASK) {
		    offset = swOverviewDispToSamp(window, x);
		    swScrollWindow(window, MAX(offset - window->length / 2, 0), /*SP_TRUE*/SP_FALSE, SP_FALSE);
		} else if (modmask & SP_EXTEND_MODIFIER_MASK) {
		    if (stdiff < 0 || (eddiff <= 0 && ABS(eddiff) > ABS(stdiff))) {
			offset = swOverviewDispToSamp(window, x);
			length = window->length - (offset - window->offset);
			swZoomRegion(window, offset, length, SP_FALSE);
			drag_type = SW_OVERVIEW_DRAG_START;
		    } else {
			offset = window->offset;
			length = swOverviewDispToSamp(window, x) - offset + 1;
			swZoomRegion(window, offset, length, SP_FALSE);
			drag_type = SW_OVERVIEW_DRAG_END;
		    }
		} else {
		    if (drag_type == SW_OVERVIEW_DRAG_NONE) {
			if (stdiff < 0) {
			    /* backward scroll */
			    swScrollWindow(window, window->offset - MAX(window->length, 1), /*SP_TRUE*/SP_FALSE, SP_FALSE);
			} else if (eddiff > 0) {
			    /* forward scroll */
			    swScrollWindow(window, window->offset + MAX(window->length, 1), /*SP_TRUE*/SP_FALSE, SP_FALSE);
			}
		    } else if (drag_type == SW_OVERVIEW_DRAG_THUMB) {
			offset_diff = swOverviewDispToSamp(window, x) - window->offset;
		    }
		}
		break;
		
	      case SP_CR_LBUTTON_MOTION:
		if (drag_type == SW_OVERVIEW_DRAG_START) {
		    offset = swOverviewDispToSamp(window, x);
		    length = window->length - (offset - window->offset);
		    swZoomRegion(window, offset, length, SP_FALSE);
		} else if (drag_type == SW_OVERVIEW_DRAG_END) {
		    offset = window->offset;
		    length = swOverviewDispToSamp(window, x) - offset + 1;
		    swZoomRegion(window, offset, length, SP_FALSE);
		} else if (drag_type == SW_OVERVIEW_DRAG_THUMB) {
		    offset = swOverviewDispToSamp(window, x) - offset_diff;
		    swScrollWindow(window, offset, /*SP_TRUE*/SP_FALSE, SP_FALSE);
		}
		break;
		
	      default:
		break;
	    }
	}
	    
	switch (reason) {
	  case SP_CR_LBUTTON_RELEASE:
	  case SP_CR_MBUTTON_RELEASE:
	  case SP_CR_RBUTTON_RELEASE:
	    if (window->data_type != SW_FREQ_DATA) {
		spDebug(50, "swOverviewButton", "drag audio end\n");
		if (drag_type == SW_OVERVIEW_DRAG_AUDIO && overview_play_start == SP_TRUE) {
		    swPlayStop(window);
		}
	    }
	    overview_play_start = SP_FALSE;
	    drag_type = SW_OVERVIEW_DRAG_NONE;
	    overview_play_start_offset = -1;
	    break;
	    
	  case SP_CR_MBUTTON_PRESS:
	  case SP_CR_RBUTTON_PRESS:
	    if (window->data_type != SW_FREQ_DATA) {
		offset = swOverviewDispToSamp(window, x);
		spDebug(50, "swOverviewButton", "drag audio started: offset = %ld\n", offset);
		if (swIsWavePlaying(window->wave) == SP_TRUE) {
		    spDebug(50, "swOverviewButton", "button press while playing\n");
		    swSetPlayStartOffset(window->wave, offset - window->wave->play_offset, SP_FALSE, SP_FALSE);
		    window->current_play_pos = offset;
		} else {
		    spDebug(50, "swOverviewButton", "button press for swPlayRegionEx\n");
		    swUpdatePoint(window, offset, SP_TRUE);
		    swPlayRegionEx(window, 0, window->wave->total_length - 1, offset);
		    overview_play_start = SP_TRUE;
		}
		drag_type = SW_OVERVIEW_DRAG_AUDIO;
		overview_play_start_offset = offset;
		spDebug(50, "swOverviewButton", "button press done\n");
	    }
	    break;
		
	  case SP_CR_MBUTTON_MOTION:
	  case SP_CR_RBUTTON_MOTION:
	    if (window->data_type != SW_FREQ_DATA && drag_type == SW_OVERVIEW_DRAG_AUDIO) {
		offset = swOverviewDispToSamp(window, x);
		spDebug(50, "swOverviewButton", "drag audio motion: offset = %ld\n", offset);
		if (swIsWavePlaying(window->wave) == SP_TRUE) {
		    if (overview_play_start_offset != offset) {
			swSetPlayStartOffset(window->wave, offset - window->wave->play_offset, SP_FALSE, SP_FALSE);
			window->current_play_pos = offset;
		    }
		} else {
		    swPlayRegionEx(window, 0, window->wave->total_length - 1, offset);
		    overview_play_start = SP_TRUE;
		}
		overview_play_start_offset = offset;
	    }
	    break;
	  default:
	    return;
	}
    }
    
    spDebug(50, "swOverviewButton", "done\n");
    
    return;
}

void swReverseOverviewRegion(swWindow window, int channel, int start_d, int end_d)
{
    if (window == NULL || window->overview_canvas == NULL || start_d < 0 || end_d < 0) return;

    spDebug(80, "swReverseOverviewRegion", "start_d = %d, end_d = %d\n", start_d, end_d);

    swFillRegion(window->overview_canvas, sw_graphics->gx_xor, window,
		 0.0, (double)window->overview_width,
		 (double)window->overview_height / (double)window->wave->num_channel,
		 channel, start_d, end_d);
    
    return;
}

void swUpdateOverview(swWindow window, spBool point_flag, spBool current_area_flag,
		      spBool selection_flag, spBool refresh_flag)
{
    int point_d = 0;
    int offset_d = 0;
    int length_d = 0;
    spGraphics graphics;
    
    if (window == NULL || window->overview_canvas == NULL) return;

    spDebug(100, "swUpdateOverview", "in: refresh_flag = %d\n", refresh_flag);
    
    /*spLockMutex(sw_graphics->mutex);*/
    
    if (point_flag == SP_TRUE && window->point >= 0) {
	point_d = swSampToOverviewDisp(window, window->point) + 1;
    }
    if (current_area_flag == SP_TRUE) {
	offset_d = swSampToOverviewDisp(window, window->offset) + 1;
	length_d = MAX(swSampToOverviewDisp(window, window->length) - 2, 1);
	
	spSetGraphicsParams(sw_graphics->gx_overview_xor,
			    SppForeground, SW_CURRENT_AREA_COLOR,
			    SppLineWidth, 2,
			    NULL);
	spDrawRectangle(window->overview_canvas, sw_graphics->gx_overview_xor,
			offset_d, 1, length_d, window->overview_height - 2);
    }
    
    if (selection_flag == SP_TRUE) {
	swReverseOverviewRegion(window, window->wave->selected_channel,
				swSampToOverviewDisp(window, window->sel_st),
				swSampToOverviewDisp(window, window->sel_ed));
    }
    
    if (point_flag == SP_TRUE && window->point >= 0) {
	spPixel pixel;
	
	if (window->config->toplevel->graphics_mode_caps & SP_GRAPHICS_MODE_CAPS_XOR) {
	    pixel = spGetForegroundPixel(sw_graphics->gx_bg) ^ spGetForegroundPixel(sw_graphics->gx_pointer);
	    graphics = sw_graphics->gx_overview_xor;
	    spSetGraphicsParams(graphics,
				SppForegroundPixel, pixel,
				SppLineWidth, 1,
				NULL);
	} else {
	    pixel = spGetForegroundPixel(sw_graphics->gx_pointer);
	    graphics = sw_graphics->gx_pointer;
	}
	spDrawLine(window->overview_canvas, graphics,
		   point_d, 0, point_d, window->overview_height);
    }

    /*spUnlockMutex(sw_graphics->mutex);*/
    
    if (refresh_flag == SP_TRUE) {
	spDebug(100, "swUpdateOverview", "call spRefreshCanvas\n");
	spRefreshCanvas(window->overview_canvas);
    }

    spDebug(100, "swUpdateOverview", "done\n");
    
    return;
}

void swDrawOverview(swWindow window, spBool refresh_flag)
{
    int n;
    double zero_offset;
    double x_factor, y_factor;
    double data_min, data_max;
    double data_height;
    double draw_width, draw_height;
    
    if (window == NULL || window->overview_canvas == NULL) {
	spDebug(30, "swDrawOverview", "canvas is not prepared\n");
	return;
    }

    spDebug(30, "swDrawOverview", "refresh_flag = %d\n", refresh_flag);
    
    spGetSize(window->overview_canvas, &(window->overview_width), &(window->overview_height));
    spDebug(30, "swDrawOverview", "width = %d, height = %d\n",
	    window->overview_width, window->overview_height);

    spFillRectangle(window->overview_canvas, sw_graphics->gx_bg,
		    0, 0, window->overview_width, window->overview_height);
    spDebug(30, "swDrawOverview", "spFillRectangle done\n");

    if (swIsWavePeakAvailable(window->wave) == SP_FALSE) {
	spDebug(30, "swDrawOverview", "peak is not available\n");
	if (refresh_flag == SP_TRUE) {
	    spDebug(100, "swDrawOverview", "call spRefreshCanvas\n");
	    spRefreshCanvas(window->overview_canvas);
	}
	return;
    }
    
    draw_width = (double)window->overview_width;
    draw_height = (double)window->overview_height / (double)window->wave->num_channel;
    draw_height = MAX(draw_height, 2);

    spDebug(30, "swDrawOverview", "peak_buf_length = %ld, draw_width = %f, draw_height = %f\n",
	    window->wave->peak_buf_length, draw_width, draw_height);
    
    data_height = swGetDataHeight(window, window->wave, &data_min, &data_max);

    x_factor = draw_width / (double)(MAX(window->wave->peak_buf_length - 1, 1));
    spDebug(30, "swDrawOverview", "data_min = %f, data_max = %f, data_height = %f, x_factor = %f\n",
	    data_min, data_max, data_height, x_factor);
    
    for (n = 0; n < window->wave->num_channel; n++) {
	spDebug(80, "swDrawOverview", "n = %d / %d\n", n, window->wave->num_channel);
#if 1
	swDrawHScale(window->overview_canvas, window, window->wave, 0.0, draw_width, draw_height,
		     n, 0, window->wave->total_length, SP_FALSE);
#endif
	swDrawVScale(window->overview_canvas, window, window->wave, 0.0, draw_width, draw_height,
		     n, data_min, data_height, &y_factor, &zero_offset, SP_FALSE, SP_FALSE, SP_FALSE);
	
	spDebug(30, "swDrawOverview", "n = %d, y_factor = %f, zero_offset = %f\n",
		n, y_factor, zero_offset);
	
	swDrawWaveform(window->overview_canvas, window, window->wave,
		       0.0, draw_width, draw_height, zero_offset,
		       0, x_factor, y_factor, n, 0, window->wave->total_length,
		       0, window->wave->peak_buf_length, SP_TRUE, SP_FALSE);
    }

    swUpdateOverview(window, SP_TRUE, SP_TRUE, SP_TRUE, refresh_flag);
    
    spDebug(30, "swDrawOverview", "done\n");
    
    return;
}

void swDrawWave(swWindow window)
{
    int width, height;
    
    if (window == NULL || spIsCreated(window->canvas) == SP_FALSE)
	return;

    spDebug(30, "swDrawWave", "%s: in\n", window->name);
    
    /*window->drawn_pos = 0;*/
    swResetAllDrawnPos(window);
    
    if (spGetSize(window->canvas, &width, &height) == SP_FALSE) {
	return;
    }

    if (window->image == NULL
	|| (width != window->width || height != window->height)) {
	window->width = width;
	window->height = height;
	swUpdateDisplayRegion(window);
	swUpdateWaveSubAreaSize(window);
    }
    
    spSetParams(window->canvas,
		SppHorizontalContentFactor, 1.0,
		SppVerticalContentFactor, 1.0,
		NULL);
    spSetParams(window->overview_canvas,
		SppHorizontalContentFactor, 1.0,
		SppVerticalContentFactor, 1.0,
		NULL);

    spDebug(30, "swDrawWave", "%s: width = %d, height = %d\n",
	    window->name, window->width, window->height);
    
    if (window->image == NULL) {
	spCreateImage(window->canvas, "waveImage",
		      window->width, window->height,
		      SppCallbackFunc, swDrawWaveImageCB,
		      SppCallbackData, window,
		      NULL);
    } else {
	swUpdateMeterWidth(window);
	
	spRedrawImage(window->image, window->width, window->height);

	window->point_d = swSampToDisp(window, window->point);
    }
	
    swDrawCursor(window, SP_FALSE);
    
    spDebug(30, "swDrawWave", "%s: done\n", window->name);
    
    return;
}

void swDrawOverviewCB(spComponent component, swWindow window)
{
    spCallbackReason reason;
    
    if (window == NULL) return;

    reason = spGetCallbackReason(component);
    
    spDebug(50, "swDrawOverviewCB", "reason = %d\n", reason);
    
    if (window->overview_canvas == NULL) {
	window->overview_canvas = component;
	spAddCallback(window->overview_canvas, SP_BUTTON_MOTION_CALLBACK |
		      SP_BUTTON_PRESS_CALLBACK | SP_BUTTON_RELEASE_CALLBACK,
		      (spCallbackFunc)swOverviewButtonCB, (void *)window);
	spAddCallback(window->overview_canvas, SP_POINTER_MOTION_CALLBACK,
		      (spCallbackFunc)swOverviewPointerMotionCB, (void *)window);
	spAddCallback(window->overview_canvas, SP_KEY_PRESS_CALLBACK,
		      (spCallbackFunc)swKeyPressCB, (void *)window);
	spAddCallback(window->overview_canvas, SP_WHEEL_CALLBACK | SP_ZOOM_CALLBACK,
		      (spCallbackFunc)swWheelCB, (void *)window);
    }

    swDrawOverview(window, SP_TRUE);
    
    spDebug(50, "swDrawOverviewCB", "done\n");
    
    return;
}

void swDrawWaveCB(spComponent component, swWindow window)
{
    spCallbackReason reason;
    spComponent menu, menu_item, sub_menu;
    
    if (window == NULL) {
	spDebug(30, "swDrawWaveCB", "window == NULL, return\n");
	return;
    }
    
    reason = spGetCallbackReason(component);
    
    spDebug(30, "swDrawWaveCB", "reason = %d\n", reason);

    if (window->canvas == NULL) {
	spDebug(30, "swDrawWaveCB", "create popup menu\n");
	window->canvas = component;
	spAddCallback(window->canvas, SP_BUTTON_MOTION_CALLBACK |
		      SP_BUTTON_PRESS_CALLBACK | SP_BUTTON_RELEASE_CALLBACK,
		      (spCallbackFunc)swButtonMotionCB, (void *)window);
	spAddCallback(window->canvas, SP_POINTER_MOTION_CALLBACK,
		      (spCallbackFunc)swMoveCursorCB, (void *)window);
	spAddCallback(window->canvas, SP_KEY_PRESS_CALLBACK,
		      (spCallbackFunc)swKeyPressCB, (void *)window);
	spAddCallback(window->canvas, SP_WHEEL_CALLBACK | SP_ZOOM_CALLBACK,
		      (spCallbackFunc)swWheelCB, (void *)window);

	menu = spCreatePopupMenu(window->canvas, "popupMenu",
				 SppHelpPath, "menu/popup_menu.html",
				 NULL);
#if defined(SW_AH_CUSTOM)
	window->toggle_ah_mode_menu = spAddCheckBoxMenuItem(menu, "toggleAHMode",
					                    SppTitle, _("SW_MENU_TOGGLE_AH_MODE_LABEL"),
							    SppGroupId, SW_AH_GROUP_ID,
							    SppSenseLevel, SW_STATE_AH_SESSION_STARTED,
							    SppCallbackFunc, swToggleAHModeCB,
							    SppCallbackData, window,
							    SppSet, SP_FALSE,
							    NULL);
	window->link_ah_file_menu = spAddCheckBoxMenuItem(menu, "linkAHFile",
							  SppTitle, _("SW_MENU_LINK_TO_AH_FILE_LABEL"),
							  SppGroupId, SW_AH_GROUP_ID,
							  SppSenseLevel, SW_STATE_AH_SESSION_STARTED,
							  SppCallbackFunc, swLinkToAHFileCB,
							  SppCallbackData, window,
							  SppSet, window->link_ah_file,
							  NULL);
	spAddMenuSeparator(menu, "popupMenuAHModeSeparator", NULL);
#endif
	menu_item = spAddMenuItem(menu, "playWave",
				  SppTitle, SW_MENU_PLAY_LABEL,
				  SppGroupId, SW_PLAY_GROUP_ID,
				  SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
				  SppCallbackFunc, swPlayRegionCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_play",
				  NULL);
	menu_item = spAddMenuItem(menu, "playWindow",
				  SppTitle, SW_MENU_PLAY_WINDOW_LABEL,
				  SppGroupId, SW_PLAY_GROUP_ID,
				  SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
				  SppCallbackFunc, swPlayWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_WINDOW_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_play_window",
				  NULL);
	menu_item = spAddMenuItem(menu, "playFile",
				  SppTitle, SW_MENU_PLAY_FILE_LABEL,
				  SppGroupId, SW_PLAY_GROUP_ID,
				  SppSenseLevel, SW_STATE_NOT_PLAY_WAVE,
				  SppCallbackFunc, swPlayFileCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_FILE_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_play_file",
				  NULL);
	menu_item = spAddMenuItem(menu, "playStop",
				  SppTitle, SW_MENU_PLAY_STOP_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_PLAY_TIME_DATA,
				  SppCallbackFunc, swPlayStopCB,
				  SppCallbackData, window,
				  SppShortcut, SW_PLAY_STOP_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_stop",
				  NULL);
	spAddMenuSeparator(menu, "popupMenuPlaySeparator", NULL);
	menu_item = spAddMenuItem(menu, "zoomIn",
				  SppTitle, SW_MENU_ZOOM_IN_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomInCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_IN_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_zoom_in",
				  NULL);
	menu_item = spAddMenuItem(menu, "zoomOut",
				  SppTitle, SW_MENU_ZOOM_OUT_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomOutCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_OUT_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_zoom_out",
				  NULL);
	menu_item = spAddMenuItem(menu, "zoomFullOut",
				  SppTitle, SW_MENU_ZOOM_FULL_OUT_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swZoomFullOutCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_FULL_OUT_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_zoom_full_out",
				  NULL);
	menu_item = spAddMenuItem(menu, "zoomRegion",
				  SppTitle, SW_MENU_ZOOM_REGION_LABEL,
				  SppGroupId, SW_PAGE_GROUP_ID,
				  SppSenseLevel, SW_STATE_SELECT_WAVE,
				  SppCallbackFunc, swZoomRegionCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ZOOM_REGION_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_zoom_region",
				  NULL);
	spAddMenuSeparator(menu, "popupMenuZoomSeparator", NULL);
	menu_item = spAddMenuItem(menu, "alignWindow",
				  SppTitle, SW_MENU_ALIGN_WINDOW_LABEL,
				  SppGroupId, SW_WINDOW_GROUP_ID,
				  SppSenseLevel, SW_STATE_STOP_SOME_WINDOWS,
				  SppCallbackFunc, swAlignWindowCB,
				  SppCallbackData, window,
				  SppShortcut, SW_ALIGN_WINDOW_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_align",
				  NULL);
	menu_item = spAddMenuItem(menu, "selectAllRegion",
				  SppTitle, SW_MENU_SELECT_ALL_REGION_LABEL,
				  SppGroupId, SW_WINDOW_GROUP_ID,
				  SppSenseLevel, SW_STATE_SOME_WINDOWS,
				  SppCallbackFunc, swSelectAllRegionCB,
				  SppCallbackData, window,
				  SppShortcut, SW_SELECT_ALL_REGION_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_select_all_region",
				  NULL);
	menu_item = spAddMenuItem(menu, "selectFromHere",
				  SppTitle, SW_MENU_SELECT_FROM_HERE_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swSelectFromHereCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_select_from_here",
				  NULL);
	menu_item = spAddMenuItem(menu, "selectToHere",
				  SppTitle, SW_MENU_SELECT_TO_HERE_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swSelectToHereCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_select_to_here",
				  NULL);

#ifdef SW_SUPPORT_CLIPBOARD
	if (swIsClipboardWindow(window) == SP_FALSE) {
	    spAddMenuSeparator(menu, "popupMenuEditSeparator", NULL);
	    sub_menu = spAddSubMenu(menu, "editPopupMenu",
				    SppTitle, SW_MENU_EDIT_LABEL,
				    SppHelpPath, "menu/popup_menu.html#popup_edit",
				    NULL);
	    menu_item = spAddMenuItem(sub_menu, "cutWave",
				      SppTitle, SW_MENU_CUT_LABEL,
				      SppGroupId, SW_CLIPBOARD_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_SELECT_TIME_CHANNELS,
				      SppCallbackFunc, swCutWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_CUT_SHORTCUT,
				      SppHelpPath, "menu/popup_menu.html#popup_cut",
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "copyWave",
				      SppTitle, SW_MENU_COPY_LABEL,
				      SppGroupId, SW_CLIPBOARD_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
				      SppCallbackFunc, swCopyWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_COPY_SHORTCUT,
				      SppHelpPath, "menu/popup_menu.html#popup_copy",
				      NULL);
	    spAddMenuSeparator(sub_menu, "popupSubMenuClipboardSeparator", NULL);
	    menu_item = spAddMenuItem(sub_menu, "paste",
				      SppTitle, SW_MENU_PASTE_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				      SppCallbackFunc, swPasteWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_PASTE_SHORTCUT,
				      SppHelpPath, "menu/popup_menu.html#popup_paste",
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "mix",
				      SppTitle, SW_MENU_MIX_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				      SppCallbackFunc, swMixWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_MIX_SHORTCUT,
				      SppHelpPath, "menu/popup_menu.html#popup_mix",
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "insert",
				      SppTitle, SW_MENU_INSERT_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD,
				      SppCallbackFunc, swInsertWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_INSERT_SHORTCUT,
				      SppHelpPath, "menu/popup_menu.html#popup_insert",
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "replace",
				      SppTitle, SW_MENU_REPLACE_LABEL,
				      SppGroupId, SW_CLIPBOARD_GROUP_ID,
				      SppSenseLevel, SW_STATE_EXIST_CLIPBOARD_SELECTED,
				      SppCallbackFunc, swReplaceWindowCB,
				      SppCallbackData, window,
				      SppShortcut, SW_REPLACE_SHORTCUT,
				      SppHelpPath, "menu/popup_menu.html#popup_replace",
				      NULL);
	    spAddMenuSeparator(sub_menu, "popupSubMenuEditSeparator", NULL);
	    menu_item = spAddMenuItem(sub_menu, "valueChange",
				      SppTitle, SW_MENU_CHANGE_VALUE_LABEL,
				      SppGroupId, SW_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_TIME_DATA,
				      SppCallbackFunc, swPopupValueChangeDialogCB,
				      SppCallbackData, window,
				      SppHelpPath, "menu/popup_menu.html#popup_change_value",
				      NULL);
	    menu_item = spAddMenuItem(sub_menu, "insertPause",
				      SppTitle, SW_MENU_INSERT_PAUSE_LABEL,
				      SppGroupId, SW_DATA_GROUP_ID,
				      SppSenseLevel, SW_STATE_TIME_DATA,
				      SppCallbackFunc, swPopupInsertPauseDialogCB,
				      SppCallbackData, window,
				      SppHelpPath, "menu/popup_menu.html#popup_insert_pause",
				      NULL);
	}
#endif
	
	spAddMenuSeparator(menu, "popupMenuLabelSeparator", NULL);
	sub_menu = spAddSubMenu(menu, "labelPopupMenu",
				SppTitle, SW_MENU_LABEL_LABEL,
				SppHelpPath, "menu/popup_menu.html#popup_label",
				NULL);
	menu_item = spAddMenuItem(sub_menu, "insertSimpleLabel",
				  SppTitle, SW_MENU_INSERT_SIMPLE_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swInsertSimpleLabelCB,
				  SppCallbackData, window,
				  SppShortcut, SW_INSERT_SIMPLE_LABEL_SHORTCUT,
				  SppHelpPath, "menu/popup_menu.html#popup_insert_simple_label",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "insertValueLabel",
				  SppTitle, SW_MENU_INSERT_VALUE_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swInsertValueLabelCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_insert_value_label",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "insertLabel",
				  SppTitle, SW_MENU_INSERT_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_EXIST_WAVE,
				  SppCallbackFunc, swInsertLabelCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_insert_label",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "changeLabel",
				  SppTitle, SW_MENU_CHANGE_LABEL_LABEL,
				  SppGroupId, SW_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppCallbackFunc, swChangeLabelCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_change_label",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "eraseLabel",
				  SppTitle, SW_MENU_ERASE_LABEL_LABEL,
				  SppGroupId, SW_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppShortcut, SW_ERASE_LABEL_SHORTCUT,
				  SppCallbackFunc, swEraseLabelCB,
				  SppHelpPath, "menu/popup_menu.html#popup_erase_label",
				  SppCallbackData, window,
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "eraseLabelRegion",
				  SppTitle, SW_MENU_ERASE_LABEL_REGION_LABEL,
				  SppGroupId, SW_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swEraseLabelRegionCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_erase_label_region",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "selectBetweenLabels",
				  SppTitle, SW_MENU_SELECT_BETWEEN_LABELS_LABEL,
				  /*SppGroupId, SW_LABEL_GROUP_ID,*/
				  SppGroupId, SW_NORMAL_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swSelectBetweenLabelsCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_select_between_labels",
				  NULL);
	spAddMenuSeparator(sub_menu, "popupMenuRegionLabelSeparator", NULL);
	menu_item = spAddMenuItem(sub_menu, "setRegionLabel",
				  SppTitle, SW_MENU_SET_REGION_LABEL_LABEL,
				  SppSenseLevel, SW_STATE_SELECT_WAVE,
				  SppShortcut, SW_SET_REGION_LABEL_SHORTCUT,
				  SppCallbackFunc, swSetRegionLabelCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_set_region_label",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "setRegionLabelAsRegion",
				  SppTitle, SW_MENU_SET_REGION_LABEL_AS_REGION_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppCallbackFunc, swSetRegionLabelAsRegionCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_set_region_label_as_region",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "catRegionLabelsRegion",
				  SppTitle, SW_MENU_CAT_REGION_LABELS_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL_HERE,
				  SppShortcut, SW_CAT_REGION_LABELS_SHORTCUT,
				  SppCallbackFunc, swCatRegionLabelsCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_concat_region_labels",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "divideRegionLabelRegion",
				  SppTitle, SW_MENU_DIVIDE_REGION_LABEL_LABEL,
				  SppGroupId, SW_REGION_LABEL_GROUP_ID,
				  SppSenseLevel, SW_STATE_EXIST_LABEL,
				  SppCallbackFunc, swDivideRegionLabelCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_divide_region_label",
				  NULL);
#ifdef SW_SUPPORT_MORPHING
	spAddMenuSeparator(menu, "popupMenuAnchorSeparator", NULL);
	swCreateAnchorSubMenu(window, menu);
#endif
	
#ifdef SW_USE_ANALYSIS
	spAddMenuSeparator(menu, "popupMenuAnalysisSeparator", NULL);
	sub_menu = spAddSubMenu(menu, "analysisPopupMenu",
				SppTitle, SW_MENU_ANALYSIS_LABEL,
				SppHelpPath, "menu/popup_menu.html#popup_analysis",
				NULL);
	menu_item = spAddMenuItem(sub_menu, "analysisRegion",
				  SppTitle, SW_MENU_ANALYSIS_REGION_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_SELECT_TIME_DATA,
				  SppShortcut, SW_ANALYSIS_REGION_SHORTCUT,
				  SppCallbackFunc, swAnalysisRegionCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_analysis_region",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "analysisWide",
				  SppTitle, SW_MENU_ANALYSIS_WIDE_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_TIME_DATA,
				  SppCallbackFunc, swAnalysisWideCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_analysis_wide",
				  NULL);
	menu_item = spAddMenuItem(sub_menu, "analysisNarrow",
				  SppTitle, SW_MENU_ANALYSIS_NARROW_LABEL,
				  SppGroupId, SW_DATA_GROUP_ID,
				  SppSenseLevel, SW_STATE_TIME_DATA,
				  SppCallbackFunc, swAnalysisNarrowCB,
				  SppCallbackData, window,
				  SppHelpPath, "menu/popup_menu.html#popup_analysis_narrow",
				  NULL);
#endif

#if 1 && defined(SW_AH_CUSTOM)
	if (window->config->toplevel->ahinforec.session_started == SP_TRUE) {
	    swUpdateAHSessionState(window);
	}
#endif
	
	swSetSenseLevel(window);
    }

    swDrawWave(window);
    
    spDebug(30, "swDrawWaveCB", "done\n");
    
    return;
}

void swRedrawWave(swWindow window)
{
    if (window == NULL) return;

    spDebug(20, "swRedrawWave", "in\n");

    /* draw wave */
    swDrawWave(window);
    
    /* set parameter value of window  */
    swSetWindowValue(window);

    spDebug(20, "swRedrawWave", "done\n");
    
    return;
}

void swReloadWave(swWindow window, swWave wave, spBool draw_subplot, spBool in_thread)
{
    spLong offset;
    spLong length;
    
    if (window == NULL) return;

    spDebug(20, "swReloadWave", "in: in_thread = %d\n", in_thread);

    if (wave == NULL) wave = window->wave;
    
    /*window->drawn_pos = 0;*/
    swResetAllDrawnPos(window);
    swSetWindowValue(window);
    
    /* read wave */
    offset = MAX(swSampToTargetSamp(window, wave, window->offset), 0);
    length = MAX(swSampToTargetSamp(window, wave, window->offset + window->length) - offset, 1);
    spDebug(20, "swReloadWave", "read: offset = %ld -> %ld, length = %ld -> %ld\n",
	    window->offset, offset, window->length, length);
    swReadWave(wave, in_thread, offset, length);

    if (draw_subplot == SP_TRUE) {
	swWaveSubArea sub_area;
	
	sub_area = swGetNextWaveSubArea(window, NULL);
		    
	while (sub_area != NULL) {
	    if (sub_area->wave != NULL && sub_area->wave != wave) {
		offset = MAX(swSampToTargetSamp(window, sub_area->wave, window->offset), 0);
		length = MAX(swSampToTargetSamp(window, sub_area->wave, window->offset + window->length) - offset, 1);
		spDebug(20, "swReloadWave", "read for subplot: offset = %ld -> %ld, length = %ld -> %ld\n",
			window->offset, offset, window->length, length);
		swReadWave(sub_area->wave, in_thread, offset, length);
	    }
	    sub_area = swGetNextWaveSubArea(window, sub_area);
	}
    }
    
    swDrawWave(window);

    spDebug(20, "swReloadWave", "done\n");

    return;
}

void swDrawAllWave(swWindow window)
{
    spComponent next = NULL;
    swWindow next_window = NULL;
    
    if (window == NULL) return;

    spDebug(100, "swDrawAllWave", "%s: in\n", window->name);
    
    swDrawWave(window);
    spDebug(100, "swDrawAllWave", "call swDrawOverview\n");
    swDrawOverview(window, SP_TRUE);
    
    next = window->window;
    while (1) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}

	if ((next_window = (swWindow)spGetUserData(next)) != NULL) {
	    spDebug(100, "swDrawAllWave", "call swDrawWave of next_window (%s)\n", next_window->name);
	    swDrawWave(next_window);
	    spDebug(100, "swDrawAllWave", "call swDrawOverview of next_window (%s)\n", next_window->name);
	    swDrawOverview(next_window, SP_TRUE);
	}
    }

    spDebug(100, "swDrawAllWave", "done\n");
    
    return;
}

void swAlignWindowCB(spComponent component, swWindow window)
{
    spLong k;
    spLong offset, length;
    double offset_f, length_f;
    spComponent next = NULL;
    swWindow next_window = NULL;

    if (window == NULL || window->wave == NULL) return;

    offset_f = swSampToDim(window, window->offset);
    length_f = swSampToDim(window, window->length);
    
    next = window->window;
    for (k = 0;; k++) {
	next = spGetNextWindow(next, SP_FALSE);
	if (next == NULL || next == window->window) {
	    break;
	}
	
	if ((next_window = (swWindow)spGetUserData(next)) != NULL
	    && next_window->wave != NULL) {
	    if (next_window->data_type == window->data_type) {
		offset = swDimToSamp(next_window, offset_f);
		length = swDimToSamp(next_window, length_f);
		length = MIN(length, next_window->wave->total_length - offset);
	    
		swZoomRegion(next_window, offset, length, SP_FALSE);
	    }
	}
    }
    
    return;
}

void swZoomRegion(swWindow window, spLong offset, spLong length, spBool in_thread)
{
    swWave wave;
    
    if (window->process_flag == SP_TRUE || window->wave == NULL) return;
    
    spDebug(50, "swZoomRegion", "in\n");

    wave = swGetTargetWave(window);
#if 1
    offset = swTargetSampToSamp(window, wave, swSampToTargetSamp(window, wave, offset));
    length = swTargetSampToSamp(window, wave, swSampToTargetSamp(window, wave, length));
#endif
    spDebug(80, "swZoomRegion", "offset = %ld, length = %ld\n", offset, length);
    
    if (offset >= 0 && length > 1 && 
	/*offset + length <= window->wave->total_length &&*/
	offset + length <= swTargetSampToSamp(window, wave, wave->total_length) &&
	!(offset == window->offset && length == window->length)) {
	swUnsetMouseCursor(window);
	
	swLockWindowMutex(window);
	window->process_flag = SP_TRUE;
	
	if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	    spDebug(100, "swZoomRegion", "call swUpdateOverview\n");
	    swUpdateOverview(window, SP_FALSE, SP_TRUE, SP_FALSE, SP_FALSE);
	}
	
	window->offset = offset;
	window->length = length;
	
	spDebug(10, "swZoomRegion", "zoom: %ld, %ld\n", window->offset, window->length);

	swReloadWave(window, wave, SP_TRUE, in_thread);
	
	if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	    swUpdateOverview(window, SP_FALSE, SP_TRUE, SP_FALSE, SP_TRUE);
	} else {
	    swDrawOverview(window, SP_TRUE);
	}
	
	window->process_flag = SP_FALSE;
	swUnlockWindowMutex(window);
    }
    
    spDebug(50, "swZoomRegion", "done\n");
    
    return;
}

void swZoomRegionCB(spComponent component, swWindow window)
{
    spLong offset, length;

    if (window != NULL && window->wave != NULL) { 
	offset = window->sel_st;
	length = window->sel_ed - window->sel_st + 1;

	swZoomRegion(window, offset, length, SP_FALSE);
    }

    return;
}

void swZoomWindow(swWindow window, double factor, spBool use_point)
{
    spLong offset, length;

    if (window != NULL && window->wave != NULL) {
	length = (spLong)spRound((double)window->length / factor);
	length = MAX(length, 2L);
	length = MIN(length, window->wave->total_length);

	if (use_point == SP_FALSE || swIsProcessing(window) == SP_TRUE) {
	    offset = window->offset + window->length / 2 - length / 2;
	} else {
	    offset = window->point - length / 2;
	}
	offset = MAX(offset, 0L);
	offset = MIN(offset, window->wave->total_length - length);

	swZoomRegion(window, offset, length, SP_FALSE);
    }

    return;
}

void swZoomInCB(spComponent component, swWindow window)
{
    swZoomWindow(window, 2.0, spIsFalse(window->pause_cursor));

    return;
}

void swZoomOutCB(spComponent component, swWindow window)
{
    swZoomWindow(window, 0.5, spIsFalse(window->pause_cursor));

    return;
}

void swZoomFullOutCB(spComponent component, swWindow window)
{
    spLong offset, length;

    if (window != NULL && window->wave != NULL) {
	length = window->wave->total_length;
	offset = 0L;

	swZoomRegion(window, offset, length, SP_FALSE);
    }

    return;
}

void swAlignAmplitudeCB(spComponent component, swWindow window)
{
    spLong k;
    double min, max;
    double range;
    swWave wave, next_wave;
    spComponent next = NULL;
    swWindow next_window = NULL;

    if (window == NULL || window->wave == NULL) return;

    wave = swGetTargetWave(window);
    
    if ((range = swGetAmplitudeRange(window, wave, &min, &max)) > 0.0) {
	if (wave->num_order > 1) {
	    min /= (double)wave->num_order;
	    max /= (double)wave->num_order;
	}
	
	next = window->window;
	for (k = 0;; k++) {
	    next = spGetNextWindow(next, SP_FALSE);
	    if (next == NULL || next == window->window) {
		break;
	    }
	
	    if ((next_window = (swWindow)spGetUserData(next)) != NULL
		&& next_window->wave != NULL) {
		next_wave = swGetTargetWave(next_window);
		
		if (next_window->data_type == window->data_type
		    && ((wave->num_order <= 1 && next_wave->num_order <= 1)
			|| (wave->num_order > 1 && next_wave->num_order > 1))) {
		    next_window->amp_min = min;
		    next_window->amp_max = max;
		    if (next_wave->num_order > 1) {
			next_window->amp_min *= (double)next_wave->num_order;
			next_window->amp_max *= (double)next_wave->num_order;
		    }
		    swUpdateVscroll(next_window);
		    swRedrawWave(next_window);
		}
	    }
	}
    }
    
    return;
}

void swZoomInAmplitudeCB(spComponent component, swWindow window)
{
    double min, max;
    double range, newrange;
    swWave wave;

    if (window != NULL && window->wave != NULL) {
	wave = swGetTargetWave(window);
	
	if ((range = swGetAmplitudeRange(window, wave, &min, &max)) > 0.0) {
	    newrange = range / 2.0;
	
	    window->amp_min = min + (range - newrange) / 2.0;
	    window->amp_max = window->amp_min + newrange;
	    swUpdateVscroll(window);
	    swRedrawWave(window);
	}
    }

    return;
}

void swZoomOutAmplitudeCB(spComponent component, swWindow window)
{
    double min, max;
    double range, newrange;
    double limit;
    swWave wave;

    if (window != NULL && window->wave != NULL) {
	wave = swGetTargetWave(window);
	
	if ((range = swGetAmplitudeRange(window, wave, &min, &max)) > 0.0) {
	    newrange = range * 2.0;

	    limit = swGetWindowLimitValue(window, wave);
	    
	    window->amp_min = min + (range - newrange) / 2.0;
	    if (limit > 1.0) {
		window->amp_min = MAX(-limit, window->amp_min);
	    }
	    window->amp_max = window->amp_min + newrange;
	    if (limit > 1.0) {
		window->amp_max = MIN(limit, window->amp_max);
	    }
	    
	    swUpdateVscroll(window);
	    swRedrawWave(window);
	}
    }

    return;
}

void swZoomFullOutAmplitudeCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL && window->amp_max > window->amp_min) {
	window->amp_min = 0.0;
	window->amp_max = -1.0;
	swUpdateVscroll(window);
	swRedrawWave(window);
    }
    
    return;
}

void swScrollAmplitudeCB(spComponent component, swWindow window)
{
    int value;
    double data_min, data_max;
    double data_range;
    double amp_range;
    swWave wave;
    
    if (window != NULL && window->wave != NULL && window->amp_max > window->amp_min) {
	if (spGetSliderValue(component, &value) == SP_FALSE) return;
	spDebug(80, "swScrollAmplitudeCB", "value = %d\n", value);

	wave = swGetTargetWave(window);
	
	data_range = swGetDataRange(window, wave, &data_min, &data_max);
	amp_range = window->amp_max - window->amp_min;
	spDebug(80, "swScrollAmplitudeCB", "data_range = %f, amp_range = %f\n", data_range, amp_range);

	window->amp_max = data_max - data_range * ((double)value / 10000.0);
	window->amp_min = window->amp_max - amp_range;
	window->amp_min = MAX(window->amp_min, data_min);
	spDebug(80, "swScrollAmplitudeCB", "updated: amp_min = %f, amp_max = %f\n",
		window->amp_min, window->amp_max);

	swRedrawWave(window);
    }

    return;
}

spLong swGetPausePlayStepTime(swWindow window, swWave wave)
{
    spLong step_time;

    step_time = swDimToSamp(window, window->config->pause_play_step_time);
    step_time = MIN(wave->total_length / 4, step_time);
    spDebug(50, "swGetPausePlayStepTime", "step_time = %ld\n", step_time);

    return step_time;
}

spBool swScrollWindowEx(swWindow window,
			int direction/* -1: prev, 0: use offset, 1: next */,
			spLong offset, spBool move_cursor, spBool disable_play_position_update,
			spBool in_thread)
{
    swWave wave;
    spLong prev_offset;
    spBool flag = SP_FALSE;

    if (window->process_flag == SP_TRUE) return SP_FALSE;
    
    spDebug(50, "swScrollWindowEx", "in_thread = %d, offset = %ld, move_cursor = %d, pause_cursor = %d\n",
	    in_thread, offset, move_cursor, window->pause_cursor);

    swUnsetMouseCursor(window);
    
    swLockWindowMutex(window);
    window->process_flag = SP_TRUE;

    wave = swGetTargetWave(window);

    prev_offset = window->offset;
    if (direction != 0 && move_cursor == SP_TRUE
	/*&& window->pause_cursor == SP_TRUE*/
	&& swIsWavePlaying(window->wave) == SP_TRUE) {
	spLong step_time;

	step_time = swGetPausePlayStepTime(window, wave);
	offset = MAX(window->current_play_pos, 0);
	
	if (direction < 0) {
	    offset -= step_time;
	} else {
	    offset += step_time;
	}
    } else {
	offset = swTargetSampToSamp(window, wave, swSampToTargetSamp(window, wave, offset));
	if (move_cursor == SP_FALSE && window->sync_play == SP_TRUE) {
	    if (window->current_play_pos >= offset && window->current_play_pos < offset + window->length) {
		disable_play_position_update = SP_TRUE;
	    }
	}
    }

    offset = MAX(offset, 0);
    if (disable_play_position_update == SP_FALSE) {
	if ((move_cursor == SP_TRUE || window->sync_play == SP_TRUE)
	    && swIsWavePlaying(window->wave) == SP_TRUE) {
	    swSetPlayStartOffset(window->wave, offset - wave->play_offset, window->sync_play, SP_FALSE);
	    window->current_play_pos = offset;
	}
    }
    offset = MIN(offset, swTargetSampToSamp(window, wave, wave->total_length) - window->length);
    
    spDebug(50, "swScrollWindowEx", "window->offset = %ld, offset = %ld\n",
	    window->offset, offset);

    if (offset != prev_offset) {
	spBool update_point = SP_FALSE;
	
	if (move_cursor == SP_TRUE
	    && swIsWaveProcessing(window->wave) == SP_FALSE
	    && swIsWavePlaying(window->wave) == SP_FALSE) {
	    update_point = SP_TRUE;
	}
	
	if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	    swUpdateOverview(window, update_point, SP_TRUE, SP_FALSE, SP_FALSE);
	}
	
	window->offset = offset;

	swReloadWave(window, wave, SP_TRUE, in_thread);
	
	if (update_point == SP_TRUE) {
	    window->point = offset + window->length / 2;
	} else {
	    move_cursor = SP_FALSE;
	}

	if (window->config->toplevel->rubber_band_selection == SP_TRUE) {
	    swUpdateOverview(window, update_point, SP_TRUE, SP_FALSE, SP_TRUE);
	} else {
	    swDrawOverview(window, SP_TRUE);
	}
    }
	
    window->process_flag = SP_FALSE;
    swUnlockWindowMutex(window);

    if (move_cursor == SP_TRUE) {
	swMoveAllCursor(window);
    }
	
    spDebug(50, "swScrollWindowEx", "done\n");
	
    return flag;
}

spBool swScrollWindow(swWindow window, spLong offset, spBool move_cursor, spBool in_thread)
{
    return swScrollWindowEx(window, 0, offset, move_cursor, SP_FALSE, in_thread);
}

void swScrollCB(spComponent component, swWindow window)
{
    int value;
    
    spDebug(80, "swScrollCB", "in\n");
	
    if (window != NULL && window->wave != NULL) {
	if (spGetSliderValue(component, &value) == SP_FALSE) return;
	spDebug(80, "swScrollCB", "value = %d\n", value);

	swScrollWindow(window, (spLong)value * window->scroll_coef,
		       spIsFalse(window->pause_cursor), SP_FALSE);
    }

    spDebug(80, "swScrollCB", "done\n");
    
    return;
}

void swForwardCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swScrollWindowEx(window, 1, window->offset + MAX(window->length / 4, 1),
			 spIsFalse(window->pause_cursor), SP_FALSE, SP_FALSE);
    }
    
    return;
}

void swBackwardCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swScrollWindowEx(window, -1, window->offset - MAX(window->length / 4, 1),
			 spIsFalse(window->pause_cursor), SP_FALSE, SP_FALSE);
    }

    return;
}

void swGoHeadCB(spComponent component, swWindow window)
{
    if (window != NULL && window->wave != NULL) {
	swScrollWindow(window, 0, spIsFalse(window->pause_cursor), SP_FALSE);
    }

    return;
}

void swGoTailCB(spComponent component, swWindow window)
{
    spLong offset;
    
    if (window != NULL && window->wave != NULL) {
	offset = window->wave->total_length - window->length;
	if (offset <= 0 && swIsWavePlaying(window->wave) == SP_TRUE) {
	    offset = window->wave->total_length - swGetPausePlayStepTime(window, window->wave);
	}
	swScrollWindow(window, offset, spIsFalse(window->pause_cursor), SP_FALSE);
    }

    return;
}

void swNextWindowCB(spComponent component, swWindow window)
{
    swWindow next_window;
    
    if ((next_window = swGetNextWindow(window)) != NULL) {
	spPopupWindow(next_window->window);
    }
    
    return;
}

void swPrevWindowCB(spComponent component, swWindow window)
{
    swWindow prev_window;

    if ((prev_window = swGetPrevWindow(window)) != NULL) {
	spPopupWindow(prev_window->window);
    }
    
    return;
}
